# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import numpy as np
import pandas as pd
from scipy.sparse.linalg import splu
from scipy.sparse import csr_matrix as sparse
from enum import Enum
from warnings import warn
from matplotlib import pyplot as plt


class DiffEqSolver(Enum):
    EULER = 1,
    RUNGE_KUTTA = 2


class DynamicModels(Enum):
    NoModel = 0,
    SynchronousGeneratorOrder4 = 1,  # fourth order synchronous machine
    SynchronousGeneratorOrder6 = 2,  # sixth order synchronous machine
    VoltageSourceConverter = 3,  # voltage source converter
    ExternalGrid = 4,  # external grid
    AsynchronousSingleCageMotor = 5,  # single cage asynchronous motor
    AsynchronousDoubleCageMotor = 6  # double cage asynchronous motor


class TransientStabilityEvents:

    def __init__(self):

        self.time = list()
        self.event_type = list()
        self.object = list()
        self.params = list()

        self.events_available = ['Bus short circuit', 'Bus recovery', 'Line failure', 'Line recovery']

    def add(self, t, evt_type, obj, param):
        """
        Add elements
        :param t: time in seconds
        :param evt_type: event type
        :param obj: object selected
        :param param: extra parameters
        """

        if evt_type not in self.events_available:
            raise Exception('Event not supported!')

        self.time.append(t)
        self.event_type.append(evt_type)
        self.object.append(obj)
        self.params.append(param)

    def remove_at(self, i):
        """
        Remove the elements at a position
        :param i: index
        """
        self.time.pop(i)
        self.event_type.pop(i)
        self.object.pop(i)
        self.params.pop(i)


class TransientStabilityResults:

    def __init__(self):

        self.name = "Transient stability"

        self.voltage = None

        self.omega = None

        self.time = None

        self.available_results = ['Bus voltage']

    def plot(self, result_type, ax=None, indices=None, names=None, LINEWIDTH=2):
        """
        Plot the results
        :param result_type:
        :param ax:
        :param indices:
        :param names:
        :return:
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        if indices is None:
            indices = np.array(range(len(names)))

        if len(indices) > 0:
            labels = names[indices]
            y_label = ''
            title = ''
            if result_type == 'Bus voltage':

                y = np.abs(self.voltage[:, indices])
                y_label = '(p.u.)'
                title = 'Bus voltage module'

            else:
                pass

            df = pd.DataFrame(data=y, columns=labels, index=self.time)

            df.fillna(0, inplace=True)

            if len(df.columns) > 10:
                df.plot(ax=ax, linewidth=LINEWIDTH, legend=False)
            else:
                df.plot(ax=ax, linewidth=LINEWIDTH, legend=True)

            ax.set_title(title)
            ax.set_ylabel(y_label)
            ax.set_xlabel('Time')

            return df

        else:
            return None


class SynchronousMachineOrder4:
    """
    4th Order Synchronous Machine Model
    https://wiki.openelectrical.org/index.php?title=Synchronous_Machine_Models#4th_Order_.28Two-Axis.29_Model
    Copyright (C) 2014-2015 Julius Susanto. All rights reserved.

    typical values:
    Ra = 0.0
    Xa = 0.0
    Xd = 1.68
    Xq = 1.61
    Xdp = 0.32
    Xqp = 0.32
    Xdpp = 0.2
    Xqpp = 0.2
    Td0p = 5.5
    Tq0p = 4.60375
    Td0pp = 0.0575
    Tq0pp = 0.0575
    H = 2
    """
    def __init__(self, H, Ra, Xd, Xdp, Xdpp, Xq, Xqp, Xqpp, Td0p, Tq0p, base_mva, Sbase, bus_idx, fn=50,
                 speed_volt=False, solver=DiffEqSolver.RUNGE_KUTTA):
        """

        :param H: is the machine inertia constant (MWs/MVA)
        :param Ra: armature resistance (pu)
        :param Xd: d-axis reactance (p.u.)
        :param Xdp: d-axis transient reactance (p.u.)
        :param Xdpp:is the d-axis subtransient reactance (pu)
        :param Xq: q-axis reactance (p.u.)
        :param Xqp: q-axis transient reactance (p.u.)
        :param Xqpp: is the q-axis subtransient reactance (pu)
        :param Td0p: d-axis transient open loop time constant (s)
        :param Tq0p: q-axis transient open loop time constant (s)
        :param base_mva: machine base power
        :param Sbase: system base power (100 MVA usually)
        :param fn: frequency
        :param speed_volt: include speed-voltage term option?
        :param solver: DiffEqSolver
        """

        self.solver = solver

        self.bus_idx = bus_idx

        self.Vfd = 0.0

        self.Id = 0.0
        self.Iq = 0.0

        # stator voltage (d, q axis)
        self.Vd = 0.0
        self.Vq = 0.0

        self.Vt = 0.0

        self.P = 0.0
        self.Q = 0.0

        self.Pm = 0.0

        self.omega = 0.0
        self.delta = 0.0
        self.Eqp = 0.0
        self.Edp = 0.0

        self.Vang = 0.0

        self.Tm = 0.0
        self.In = 0.0
        self.Im = 0.0

        self.omega_prev = 0.0
        self.delta_prev = 0.0
        self.Eqp_prev = 0.0
        self.Edp_prev = 0.0

        self.arr_Eqp = np.zeros(3)
        self.arr_Edp = np.zeros(3)
        self.arr_omega = np.zeros(3)
        self.arr_delta = np.zeros(3)

        self.Td0p = Td0p
        self.Tq0p = Tq0p

        # angular speed (w = 2·pi·f)
        self.omega_n = 2.0 * np.ones_like(H) * np.pi * fn

        # Check for speed-voltage term option
        self.speed_volt = speed_volt  # True / False

        # Convert impedances and H to system MVA base
        self.H = H * base_mva / Sbase
        self.Ra = Ra * Sbase / base_mva
        self.Xd = Xd * Sbase / base_mva
        self.Xdp = Xdp * Sbase / base_mva
        self.Xdpp = Xdpp * Sbase / base_mva
        self.Xq = Xq * Sbase / base_mva
        self.Xqp = Xqp * Sbase / base_mva
        self.Xqpp = Xqpp * Sbase / base_mva

        # Equivalent Norton impedance for Ybus
        self.Yg = self.get_yg()

    def get_yg(self):
        """
        Get the generator admittance
        :return: shunt admittance
        """
        return (self.Ra - 1j * 0.5 * (self.Xdp + self.Xqp)) / (self.Ra ** 2.0 + (self.Xdp * self.Xqp))

    def initialise(self, vt0, S0):
        """
        Initialise machine signals and states based on load flow voltage and complex power injection
        :param vt0: complex initial voltage
        :param S0: complex initial power
        :return:
        """

        # Calculate initial armature current
        Ia0 = np.conj(S0 / vt0)
        phi0 = np.angle(Ia0)

        # Calculate steady state machine emf (i.e. voltage behind synchronous reactance)
        Eq0 = vt0 + (self.Ra + 1j * self.Xq) * Ia0
        self.delta = np.angle(Eq0)

        # Convert currents to rotor reference frame
        self.Id = np.abs(Ia0) * np.sin(self.delta - phi0)
        self.Iq = np.abs(Ia0) * np.cos(self.delta - phi0)

        # Convert voltages to rotor reference frame
        self.Vd = np.abs(vt0) * np.sin(self.delta - np.angle(vt0))
        self.Vq = np.abs(vt0) * np.cos(self.delta - np.angle(vt0))

        # Calculate machine state variables and Vfd
        self.Eqp = self.Vq + self.Ra * self.Iq + self.Xdp * self.Id
        self.Edp = self.Vd + self.Ra * self.Id - self.Xqp * self.Iq
        self.Vfd = np.abs(self.Eqp) + (self.Xd - self.Xdp) * self.Id

        # Calculate active and reactive power
        self.P = (self.Vd + self.Ra * self.Id) * self.Id + (self.Vq + self.Ra * self.Iq) * self.Iq
        self.Q = self.Vq * self.Id - self.Vd * self.Iq

        # Initialise the rest
        self.Vt = np.abs(vt0)
        self.Pm = self.P
        self.omega = 1

        self.check_diffs()

    def calc_currents(self, Vbus, Ibus):
        """
        Calculate machine current injections (in network reference frame)
        :param vt: complex initial voltage
        :return:
        """

        vt = Vbus[self.bus_idx]

        # Calculate terminal voltage in dq reference frame
        self.Vd = np.abs(vt) * np.sin(self.delta - np.angle(vt))
        self.Vq = np.abs(vt) * np.cos(self.delta - np.angle(vt))

        # Check if speed-voltage term should be included
        k = np.where(self.speed_volt == False)[0]
        omega = self.omega
        omega[k] = np.ones_like(k)

        # Calculate Id and Iq (Norton equivalent current injection in dq frame)
        self.Id = (self.Eqp - self.Ra / (self.Xqp * omega) * (self.Vd - self.Edp) - self.Vq / omega) / (self.Xdp + self.Ra ** 2 / (omega * omega * self.Xqp))
        self.Iq = (self.Vd / omega + self.Ra * self.Id / omega - self.Edp) / self.Xqp

        # Calculate power output
        self.P = (self.Vd + self.Ra * self.Id) * self.Id + (self.Vq + self.Ra * self.Iq) * self.Iq
        self.Q = self.Vq * self.Id - self.Vd * self.Iq

        # Calculate machine current injection (Norton equivalent current injection in network frame)
        self.In = (self.Iq - 1j * self.Id) * np.exp(1j * self.delta)
        self.Im = self.In + self.Yg * vt

        self.Vt = np.abs(vt)
        self.Vang = np.angle(vt)

        # apply th currents to the passed vector
        Ibus[self.bus_idx] = self.Im

    def check_diffs(self):
        """
        Check if differential equations are zero (on initialisation)
        """

        # State variables
        dEqp = (self.Vfd - (self.Xd - self.Xdp) * self.Id - self.Eqp) / self.Td0p
        dEdp = ((self.Xq - self.Xqp) * self.Iq - self.Edp) / self.Tq0p

        if np.round(dEdp, 6).all() != 0 or np.round(dEqp, 6).all() != 0:
            warn('Warning: differential equations not zero on initialisation...')
            print('dEdp = ' + str(dEdp) + ', dEqp = ' + str(dEqp))

    def function(self, h, Eqp, Edp, omega):
        """
        Compute the magnitude's derivatives
        :param Eqp:
        :param Edp:
        :param omega:
        :return:
        """
        # Electrical differential equations
        f1 = (self.Vfd - (self.Xd - self.Xdp) * self.Id - Eqp) / self.Td0p

        f2 = ((self.Xq - self.Xqp) * self.Iq - Edp) / self.Tq0p

        # Swing equation
        f3 = 1.0 / (2.0 * self.H) * (self.Pm / omega - self.P)

        f4 = self.omega_n * (omega - 1.0)

        return h * f1, h * f2, h * f3, h * f4

    def solve(self, h):
        """
        Solve using Runge-Kutta
        Args:
            h: step size

        Returns: self.Eqp, self.Edp, self.omega, self.delta
        """
        # step 1
        k1_Eqp, k1_Edp, k1_omega, k1_delta = self.function(h, self.Eqp, self.Edp, self.omega)

        # step 2
        k2_Eqp, k2_Edp, k2_omega, k2_delta = self.function(h, self.Eqp + 0.5 * k1_Eqp,
                                                           self.Edp + 0.5 * k1_Edp,
                                                           self.omega + 0.5 * k1_omega)

        # step 3
        k3_Eqp, k3_Edp, k3_omega, k3_delta = self.function(h, self.Eqp + 0.5 * k2_Eqp,
                                                           self.Edp + 0.5 * k2_Edp,
                                                           self.omega + 0.5 * k2_omega)

        # step 4
        k4_Eqp, k4_Edp, k4_omega, k4_delta = self.function(h, self.Eqp + 0.5 * k3_Eqp,
                                                           self.Edp + 0.5 * k3_Edp,
                                                           self.omega + 0.5 * k3_omega)
        a = 1.0 / 6.0

        # Tm = Pm / omega_0

        # update the values

        self.Eqp += a * (k1_Eqp + 2.0 * k2_Eqp + 2.0 * k3_Eqp + k4_Eqp)

        self.Edp += a * (k1_Edp + 2.0 * k2_Edp + 2.0 * k3_Edp + k4_Edp)

        self.omega += a * (k1_omega + 2.0 * k2_omega + 2.0 * k3_omega + k4_omega)

        self.delta += a * (k1_delta + 2.0 * k2_delta + 2.0 * k3_delta + k4_delta)

        return self.Eqp, self.Edp, self.omega, self.delta


class SynchronousMachineOrder6SauerPai:
    """
    PYPOWER-Dynamics
    6th Order Synchronous Machine Model
    Based on Sauer-Pai model
    Sauer, P.W., Pai, M. A., "Power System Dynamics and Stability", Stipes Publishing, 2006 
    """
    
    def __init__(self, H, Ra, Xa, Xd, Xdp, Xdpp, Xq, Xqp, Xqpp, Td0p, Tq0p, Td0pp, Tq0pp, base_mva, Sbase, bus_idx,
                 fn=50, speed_volt=False):
        
        self.omega_n = 2 * np.pi * fn

        self.bus_idx = bus_idx

        # Check for speed-voltage term option 
        self.speed_volt = speed_volt

        self.H = H * base_mva / Sbase
        self.Ra = Ra * Sbase / base_mva
        self.Xa = Xa * Sbase / base_mva
        self.Xd = Xd * Sbase / base_mva
        self.Xdp = Xdp * Sbase / base_mva
        self.Xdpp = Xdpp * Sbase / base_mva
        self.Xq = Xq * Sbase / base_mva
        self.Xqp = Xqp * Sbase / base_mva
        self.Xqpp = Xqpp * Sbase / base_mva

        self.Td0p, self.Tq0p, self.Td0pp, self.Tq0pp = Td0p, Tq0p, Td0pp, Tq0pp

        # Internal variables
        self.gamma_d1 = (self.Xdpp - self.Xa) / (self.Xdp - self.Xa)
        self.gamma_d2 = (1 - self.gamma_d1) / (self.Xdp - self.Xa)
        self.gamma_q1 = (self.Xqpp - self.Xa) / (self.Xqp - self.Xa)
        self.gamma_q2 = (1 - self.gamma_q1) / (self.Xqp - self.Xa)

        # Equivalent Norton impedance for Ybus modification
        self.Yg = (self.Ra - 1j * 0.5 * (self.Xdpp + self.Xqpp)) / (self.Ra** 2 + (self.Xdpp * self.Xqpp))

        # results
        self.Vfd = 0
        self.Id = 0
        self.Iq = 0
        self.Vd = 0
        self.Vq = 0
        self.Vt = 0
        self.Vang = 0
        self.P = 0
        self.Q = 0
        self.Pm = 0
        self.Tm = 0
        self.omega = 1
        self.delta = 0
        self.Eqp = 0
        self.phiq_pp = 0
        self.Edp = 0
        self.phid_pp = 0
        self.In = 0
        self.Im = 0

    def get_yg(self):
        """
        Get the generator admittance
        :return: shunt admittance
        """
        return (self.Ra - 1j * 0.5 * (self.Xdp + self.Xqp)) / (self.Ra ** 2.0 + (self.Xdp * self.Xqp))

    def initialise(self, vt0, S0):
        """
        Initialise machine signals and states based on load flow voltage and complex power injection
        """

        # Calculate initial armature current
        Ia0 = np.conj(S0 / vt0)
        phi0 = np.angle(Ia0)

        # Calculate steady state machine emf (i.e. voltage behind synchronous reactance)
        Eq0 = vt0 + (self.Ra + 1j * self.Xq) * Ia0
        self.delta = np.angle(Eq0)

        # Convert currents to rotor reference frame
        self.Id = np.abs(Ia0) * np.sin(self.delta - phi0)
        self.Iq = np.abs(Ia0) * np.cos(self.delta - phi0)

        self.Vd = np.abs(vt0) * np.sin(self.delta - np.angle(vt0))
        self.Vq = np.abs(vt0) * np.cos(self.delta - np.angle(vt0))

        # Calculate machine state variables and Vfd
        self.Edp = self.Vd - self.Xqpp * self.Iq + self.Ra * self.Id - (1 - self.gamma_q1) * (self.Xqp - self.Xa) * self.Iq
        self.Eqp = self.Vq + self.Xdpp * self.Id + self.Ra * self.Iq + (1 - self.gamma_d1) * (self.Xdp - self.Xa) * self.Id
        self.phid_pp = self.Eqp - (self.Xdp - self.Xa) * self.Id
        self.phiq_pp = -self.Edp - (self.Xqp - self.Xa) * self.Iq
        self.Vfd = self.Eqp + (self.Xd - self.Xdp) * (self.Id - self.gamma_d2 * self.phid_pp - (1 - self.gamma_d1) * self.Id + self.gamma_d2 * self.Eqp)

        # Calculate active and reactive power
        self.P = self.Vd * self.Id + self.Vq * self.Iq
        self.Q = self.Vq * self.Id - self.Vd * self.Iq

        # Initialise signals, states and parameters
        self.Vt = np.abs(vt0)
        self.Vang = 0
        self.Pm = self.P
        self.Tm = self.P
        self.omega = 1

        self.check_diffs()

    def check_diffs(self):
        """
        Check if differential equations are zero (on initialisation)
        """

        dEqp = (self.Vfd - (self.Xd - self.Xdp)
                * (self.Id - self.gamma_d2 * self.phid_pp - (1 - self.gamma_d1)
                * self.Id + self.gamma_d2 * self.Eqp) - self.Eqp) / self.Td0p
        dEdp = ((self.Xq - self.Xqp)
                * (self.Iq - self.gamma_q2 * self.phiq_pp - (1 - self.gamma_q1)
                * self.Iq - self.gamma_q2 * self.Edp) - self.Edp) / self.Tq0p
        dphid_pp = (self.Eqp - (self.Xdp - self.Xa) * self.Id - self.phid_pp) / self.Td0pp
        dphiq_pp = (-self.Edp - (self.Xqp - self.Xa) * self.Iq - self.phiq_pp) / self.Tq0pp

        if np.round(dEdp, 6).all() != 0 or np.round(dEqp, 6).all() != 0 or \
                        np.round(dphid_pp, 6).all() != 0 or np.round(dphiq_pp, 6).all() != 0:
            print('Warning: differential equations not zero on initialisation...')
            print('dEdp = ' + str(dEdp) + ', dEqp = ' + str(dEqp) + ', dphid_pp = ' + str(
                dphid_pp) + ', dphiq_pp = ' + str(dphiq_pp))

    def calc_currents(self, vt):
        """
        Calculate machine current injections (in network reference frame)
        """

        # Calculate terminal voltage in dq reference frame
        self.Vd = np.abs(vt) * np.sin(self.delta - np.angle(vt))
        self.Vq = np.abs(vt) * np.cos(self.delta - np.angle(vt))

        # Check if speed-voltage term should be included
        if self.speed_volt:
            omega = self.omega
        else:
            omega = 1

        # Calculate Id and Iq (Norton equivalent current injection in dq frame)
        self.Id = (-self.Vq / omega + self.gamma_d1 * self.Eqp + (1 - self.gamma_d1) * self.phid_pp
                   - self.Ra / (omega * self.Xqpp)
                    * (self.Vd - self.gamma_q1 * self.Edp + (1 - self.gamma_q1) * self.phiq_pp)) \
                    / (self.Xdpp + self.Ra**2 / (omega * omega * self.Xqpp))

        self.Iq = (self.Vd / omega + (self.Ra * self.Id / omega) - self.gamma_q1 * self.Edp
                   + (1 - self.gamma_q1) * self.phiq_pp) / self.Xqpp

        # Calculate power output
        self.P = (self.Vd + self.Ra * self.Id) * self.Id + (self.Vq + self.Ra * self.Iq) * self.Iq
        self.Q = self.Vq * self.Id - self.Vd * self.Iq
        # self.S = np.complex(p, q)

        # Calculate machine current injection (Norton equivalent current injection in network frame)
        delta = self.delta
        self.In = (self.Iq - 1j * self.Id) * np.exp(1j * self.delta)
        self.Im = self.In + self.Yg * vt

        # Update signals
        self.Vt = np.abs(vt)
        self.Vang = np.angle(vt)

        return self.Im

    def function(self, Eqp, Edp, omega):
        """
        Solve machine differential equations for the next stage in the integration step
        :param Eqp:
        :param Edp:
        :param omega:
        :return:
        """

        # Eq'
        f1 = (self.Vfd
              - (self.Xd - self.Xdp) * (self.Id - self.gamma_d2 * self.phid_pp
                - (1 - self.gamma_d1) * self.Id + self.gamma_d2 * Eqp) - Eqp) / self.Td0p

        # Ed'
        f2 = ((self.Xq - self.Xqp) * (self.Iq - self.gamma_q2 * self.phiq_pp
              - (1 - self.gamma_q1) * self.Iq - self.gamma_q2 * Edp) - Edp) / self.Tq0p

        # phi d pp
        f3 = (Eqp - (self.Xdp - self.Xa) * self.Id - self.phid_pp) / self.Td0pp

        # phi_q''
        f4 = (-Edp - (self.Xqp - self.Xa) * self.Iq - self.phiq_pp) / self.Tq0pp

        # omega
        f5 = 0.5 / self.H * (self.Pm / omega - self.P)

        # delta
        f6 = self.omega_n * (omega - 1)

        return f1, f2, f3, f4, f5, f6


class VoltageSourceConverterAverage:
    """
    Voltage Source Converter Model Class
    Average model of a VSC in voltage-control mode (i.e. controlled voltage source behind an impedance).
    Copyright (C) 2014-2015 Julius Susanto. All rights reserved.
    """
    def __init__(self, Rl, Xl, fn, bus_idx):

        self.bus_idx = bus_idx

        self.Rl = Rl
        self.Xl = Xl
        self.fn = fn

        self.Edq = 0.0
        self.delta = 0.0
        self.omega = 0.0
        self.Id = 0.0
        self.Iq = 0.0
        self.Vt = 0.0
        self.Ed = 0.0
        self.Eq = 0.0
        self.Vd = 0.0
        self.Vq = 0.0

        self.In = 0.0
        self.Im = 0.0

        # Equivalent Norton impedance for Ybus modification
        self.Yg = self.get_yg()

    def get_yg(self):
        """
        Get the generator admittance
        :return: shunt admittance
        """
        return 1 / (self.Rl + 1j * self.Xl)

    def initialise(self, vt0, S0):
        """
        Initialise converter emf based on load flow voltage and grid current injection
        :param vt0: complex voltage
        :param S0: complex power
        """
        # Calculate initial armature current
        Ia0 = np.conj(S0 / vt0)
        phi0 = np.angle(Ia0)

        # Calculate steady state machine emf (i.e. voltage behind synchronous reactance)
        self.Edq = vt0 + (self.Rl + 1j * self.Xl) * Ia0
        self.delta = np.angle(self.Edq)

        # Convert currents to rotor reference frame
        self.Id = np.abs(Ia0) * np.sin(self.delta - phi0)
        self.Iq = np.abs(Ia0) * np.cos(self.delta - phi0)

        # Initialise signals, states and parameters
        self.Vt = np.abs(vt0)
        self.Ed = np.real(self.Edq)
        self.Eq = np.imag(self.Edq)
        self.omega = 1.0

    def calc_currents(self, vt):
        """
        Solve grid current injections (in network reference frame)
        :param vt: complex voltage
        :return:
        """
        self.Edq = self.Ed + 1j * self.Eq
        self.delta = np.angle(self.Edq)

        # Calculate terminal voltage in dq reference frame
        self.Vd = np.abs(vt) * np.sin(self.delta - np.angle(vt))
        self.Vq = np.abs(vt) * np.cos(self.delta - np.angle(vt))

        # Calculate Id and Iq (Norton equivalent current injection in dq frame)
        Ia = (self.Edq - vt) / (self.Rl + 1j * self.Xl)
        phi = np.angle(Ia)
        self.Id = np.abs(Ia) * np.sin(self.delta - phi)
        self.Iq = np.abs(Ia) * np.cos(self.delta - phi)

        # Calculate machine current injection (Norton equivalent current injection in network frame)
        self.In = (self.Iq - 1j * self.Id) * np.exp(1j * self.delta)
        self.Im = self.In + self.Yg * vt
        self.Vt = np.abs(vt)

        # self.delta = delta

        return self.Im

    def function(self, h, d):
        """
        Solve machine differential equations for the next stage in the integration step
        :param h: solve step in seconds
        """

        # State variables do not change in this model
        pass


class ExternalGrid:
    """
    External Grid Model Class
    Grid is modelled as a constant voltage behind a transient reactance
    and two differential equations representing the swing equations.
    """
    def __init__(self, Xdp, H, fn, bus_idx):

        self.bus_idx = bus_idx

        self.Xdp = Xdp
        self.H = H
        self.fn = fn

        # result values
        self.Vt = 0
        self.P = 0
        self.Pm = 0
        self.Eq = 0
        self.omega = 0
        self.delta = 0

    def initialise(self, vt0, S0):
        """
        Initialise grid emf based on load flow voltage and grid current injection
        """
        # Calculate initial armature current
        Ia0 = np.conj(S0 / vt0)
        phi0 = np.angle(Ia0)

        # Calculate steady state machine emf (i.e. voltage behind synchronous reactance)
        Eq0 = vt0 + 1j * self.Xdp * Ia0
        delta0 = np.angle(Eq0)

        p0 = 1 / self.Xdp * np.abs(vt0) * np.abs(Eq0) * np.sin(delta0 - np.angle(vt0))

        # Initialise signals, states and parameters
        self.Vt = np.abs(vt0)
        self.P = p0
        self.Pm = p0
        self.Eq = np.abs(Eq0)
        self.omega = 1
        self.delta = delta0

    def get_yg(self):
        """
        Return the shunt admittance
        Returns:

        """
        return 1 / (1j * self.Xdp)

    def calc_currents(self, vt):
        """
        Solve grid current injections (in network reference frame)
        """

        self.P = np.abs(vt) * self.Eq * np.sin(self.delta - np.angle(vt)) / self.Xdp

        # Update signals
        self.Vt = np.abs(vt)

        i_grid = self.Eq * np.exp(1j * self.delta) / (1j * self.Xdp)

        return i_grid

    def function(self, Eqp, Edp, omega):
        """
        Solve machine differential equations for the next stage in the integration step
        """

        # Solve swing equation
        f1 = 1 / (2 * self.H) * (self.Pm / omega - self.P)

        f2 = 2 * np.pi * self.fn * (omega - 1.0)

        f3 = np.zeros_like(Eqp)

        f4 = np.zeros_like(Eqp)

        return f1, f2, f3, f4


class SingleCageAsynchronousMotor:
    """
    Single Cage Asynchronous Motor Model

    Model equations based on section 15.2.4 of:
    Milano, F., "Power System Modelling and Scripting", Springer-Verlag, 2010

    """

    def __init__(self, H, Rr, Xr, Rs, Xs, a, Xm, MVA_Rating, Sbase, bus_idx, fn=50):
        """
        
        :param H: 
        :param Rr: 
        :param Xr: 
        :param Rs: 
        :param Xs: 
        :param a: 
        :param Xm: 
        :param Sbase: System base power
        :param fn: system frequency
        """

        self.bus_idx = bus_idx

        self.omega_n = 2 * np.pi * fn

        self.Sbase = Sbase
        self.base_mva = MVA_Rating
        # Convert parameters to 100MVA base
        self.H = H * self.base_mva / 100
        self.a = a * self.base_mva / 100
        self.Rs = Rs * 100 / self.base_mva
        self.Xs = Xs * 100 / self.base_mva
        self.Xm = Xm * 100 / self.base_mva
        self.Rr = Rr * 100 / self.base_mva
        self.Xr = Xr * 100 / self.base_mva

        # Calculate internal parameters
        self.X0 = self.Xs + self.Xm
        self.Xp = self.Xs + self.Xr * self.Xm / (self.Xr + self.Xm)
        self.T0p = (self.Xr + self.Xm) / (self.omega_n * self.Rr)

        # Motor start signal
        self.start = 0

        # Equivalent Norton impedance for Ybus modification (NOTE: currently not used)
        self.Ym = self.Rs - 1j * self.Xs

        # results
        self.Id = np.zeros_like(H)
        self.Iq = np.zeros_like(H)
        self.Vd = np.zeros_like(H)
        self.Vq = np.zeros_like(H)
        self.Vt = np.zeros_like(H)
        self.P = np.zeros_like(H)
        self.Q = np.zeros_like(H)
        self.Te = np.zeros_like(H)
        self.slip = np.zeros_like(H)
        self.Eqp = np.zeros_like(H)
        self.Edp = np.zeros_like(H)
        self.omega = 1 - self.slip
        self.Im = np.zeros_like(H)
        self.In = np.zeros_like(H)
        self.Vang = np.zeros_like(H)

    def get_yg(self):
        """
        Get the generator admittance
        :return: shunt admittance
        """
        return 1 / (self.Rr + 1j * self.Xr)

    def initialise(self, vt0, S0):
        """
        Initialise machine signals and states based on load flow voltage and complex power injection
        NOTE: currently only initialised at standstill
        """
     
        # Initialise signals, states and parameters
        self.Id = 0
        self.Iq = 0
        self.Vd = 0
        self.Vq = 0
        self.Vt = np.abs(vt0)
        self.P = 0
        self.Q = 0
        self.Te = 0
        self.omega = 1 - 0

        self.slip = 1
        self.Eqp = 0
        self.Edp = 0

        self.check_diffs()

    def calc_tmech(self, s):
        """
        Calculate mechanical load torque (with a quadratic load model)
        :param s: slip
        :return: 
        """

        Tm = self.a * (1 - s) ** 2

        return Tm

    def calc_currents(self, vt):
        """
        Calculate machine current injections (in network reference frame)
        """

        if self.start == 1:
            # Calculate terminal voltage in dq reference frame (set to rotate with q-axis)
            self.Vd = -np.abs(vt) * np.sin(np.angle(vt))
            self.Vq = np.abs(vt) * np.cos(np.angle(vt))

            # Calculate Id and Iq (Norton equivalent current injection in dq frame)
            self.Iq = (self.Rs / self.Xp * (self.Vq - self.Eqp) - self.Vd + self.Edp) / (self.Xp + self.Rs ** 2 / self.Xp)
            self.Id = (self.Vq - self.Eqp - self.Rs * self.Iq) / self.Xp

            # Calculate power output and electrical torque
            self.P = -(self.Vd * self.Id + self.Vq * self.Iq)
            self.Q = -(self.Vq * self.Id - self.Vd * self.Iq)
            self.Te = (self.Edp * self.Id + self.Eqp * self.Iq)  # / self.omega_n

            # Calculate machine current injection (Norton equivalent current injection in network frame)
            self.In = (self.Id + 1j * self.Iq) * np.exp(1j * (-np.pi / 2))
            self.Im = -self.In  # + self.Ym * vt

            # Update signals
            self.Vt = np.abs(vt)
            self.Vang = np.angle(vt)
            self.omega = 1 - self.slip

        else:
            self.Im = 0

        return self.Im

    def function(self, Edp, Eqp):
        """
        Solve machine differential equations for the next stage in the integration step
        """

        if self.start == 1:

            # Eq'
            f1 = (-self.omega_n * self.slip * Edp - (Eqp - (self.X0 - self.Xp) * self.Id) / self.T0p) * self.base_mva / self.Sbase

            # Ed'
            f2 = (self.omega_n * self.slip * Eqp - (Edp + (self.X0 - self.Xp) * self.Iq) / self.T0p) * self.base_mva / self.Sbase

            # Tm
            Tm = self.calc_tmech(self.slip)
            f3 = (Tm - self.Te) / (2 * self.H)
        
        else:
            f1 = np.zeros_like(Edp)
            f2 = np.zeros_like(Edp)
            f3 = np.zeros_like(Edp)
        
        return f1, f2, f3

    def check_diffs(self):
        """
        Check if differential equations are zero (on initialisation)
        """

        # State variables
        dEdp = self.omega_n * self.slip * self.Eqp - (self.Edp + (self.X0 - self.Xp) * self.Iq) / self.T0p
        dEqp = -self.omega_n * self.slip * self.Edp - (self.Eqp - (self.X0 - self.Xp) * self.Id) / self.T0p
        ds = self.calc_tmech(1) - self.Te

        if np.round(dEdp, 6).all() != 0 or np.round(dEqp, 6).all() != 0 or np.round(ds, 6).all() != 0:
            warn('Warning: differential equations not zero on initialisation...')
            print('dEdp = ', dEdp, ', dEqp = ', dEqp, ', ds = ', ds)


class DoubleCageAsynchronousMotor:

    """
    Double Cage Asynchronous Machine Model

    Model equations based on section 15.2.5 of:
    Milano, F., "Power System Modelling and Scripting", Springer-Verlag, 2010

    """

    def __init__(self, H, Rr, Xr, Rs, Xs, a, Xm, Rr2, Xr2, MVA_Rating, Sbase, bus_idx, fn=50):

        self.bus_idx = bus_idx

        self.omega_n = 2 * np.pi * fn

        # Convert parameters to 100MVA base
        self.base_mva = MVA_Rating
        self.Sbase = Sbase
        self.H = H * self.base_mva / self.Sbase
        self.a = a * self.base_mva / self.Sbase
        self.Rs = Rs * self.Sbase / self.base_mva
        self.Xs = Xs * self.Sbase / self.base_mva
        self.Xm = Xm * self.Sbase / self.base_mva
        self.Rr = Rr * self.Sbase / self.base_mva
        self.Xr = Xr * self.Sbase / self.base_mva
        self.Rr2 = Rr2 * self.Sbase / self.base_mva
        self.Xr2 = Xr2 * self.Sbase / self.base_mva

        # Calculate internal parameters
        self.X0 = self.Xs + self.Xm
        self.Xp = self.Xs + self.Xr * self.Xm / (self.Xr + self.Xm)
        self.Xpp = self.Xs + self.Xr * self.Xr2 * self.Xm / (self.Xr * self.Xr2 + self.Xm * self.Xr + self.Xm * self.Xr2)
        self.T0p = (self.Xr + self.Xm) / (self.omega_n * self.Rr)
        self.T0pp = (self.Xr2 + (self.Xr * self.Xm) / (self.Xr + self.Xm)) / (self.omega_n * self.Rr2)

        # Motor start signal
        self.start = 0

        # Equivalent Norton impedance for Ybus modification (NOTE: currently not used)
        self.Ym = self.Rs - 1j * self.Xs

        # results
        self.Id = 0
        self.Iq = 0
        self.Vd = 0
        self.Vq = 0
        self.Vt = 0
        self.P = 0
        self.Q = 0
        self.Te = 0
        self.slip = 0
        self.Eqp = 0
        self.Edp = 0
        self.Eqpp = 0
        self.Edpp = 0
        self.omega = 1 - self.slip
        self.Im = 0.0
        self.In = 0.0
        self.Vang = 0.0

    def get_yg(self):
        """
        Get the generator admittance
        :return: shunt admittance
        """
        return 1 / (self.Rr + 1j * self.Xr)

    def initialise(self, vt0, S0):
        """
        Initialise machine signals and states based on load flow voltage and complex power injection
        NOTE: currently only initialised at standstill
        """

        # Initialise signals, states and parameters
        self.Id = 0
        self.Iq = 0
        self.Vd = 0
        self.Vq = 0
        self.Vt = np.abs(vt0)
        self.P = 0
        self.Q = 0
        self.Te = 0
        self.slip = 1
        self.omega = 1 - self.slip
        
        self.Eqp = 0
        self.Edp = 0
        self.Eqpp = 0
        self.Edpp = 0

        self.check_diffs()

    def calc_tmech(self, s):
        """
        Calculate mechanical load torque (with a quadratic load model)
        """

        Tm = self.a * (1 - s) ** 2

        return Tm

    def calc_currents(self, vt):
        """
        Calculate machine current injections (in network reference frame)
        """

        if self.start == 1:
            # Calculate terminal voltage in dq reference frame (set to rotate with q-axis)
            self.Vd = -np.abs(vt) * np.sin(np.angle(vt))
            self.Vq = np.abs(vt) * np.cos(np.angle(vt))

            # Calculate Id and Iq (Norton equivalent current injection in dq frame)
            self.Iq = (self.Rs / self.Xpp * (self.Vq - self.Eqpp) - self.Vd + self.Edpp) / (self.Xpp + self.Rs ** 2 / self.Xpp)
            self.Id = (self.Vq - self.Eqpp - self.Rs * self.Iq) / self.Xpp

            # Calculate power output and electrical torque
            self.P = -(self.Vd * self.Id + self.Vq * self.Iq)
            self.Q = -(self.Vq * self.Id -self. Vd * self.Iq)
            self.Te = self.Edpp * self.Id + self.Eqpp * self.Iq

            # Calculate machine current injection (Norton equivalent current injection in network frame)
            self.In = (self.Id + 1j * self.Iq) * np.exp(1j * (-np.pi / 2))
            self.Im = -self.In  # + self.Ym * vt

            # Update signals
            self.Vt = np.abs(vt)
            self.Vang = np.angle(vt)
            self.omega = 1 - self.slip

        else:
            self.Im = 0

        return self.Im

    def solve_step(self, Edp, Eqp, Edpp, Eqpp):
        """
        Solve machine differential equations for the next stage in the integration step
        """

        if self.start == 1:

            # Eq'
            f1 = (-self.omega_n * self.slip * Edp - (Eqp - (self.X0 - self.Xp) * self.Id) / self.T0p) * self.base_mva / self.Sbase
            # k_Eqp = h * f1
            
            # Ed'
            f2 = (self.omega_n * self.slip * Eqp - (Edp + (self.X0 - self.Xp) * self.Iq) / self.T0p) * self.base_mva / self.Sbase
            # k_Edp = h * f2
            
            # Eq''
            f3 = f1 + (self.omega_n * self.slip * (Edp - Edpp) + (
                        Eqp - Eqpp + (self.Xp - self.Xpp) * self.Id) / self.T0pp) * self.base_mva / self.Sbase
            # k_Eqpp = h * f3
            
            # Ed''
            f4 = f2 + (-self.omega_n * self.slip * (Eqp - Eqpp) + (
                        Edp - Edpp - (self.Xp - self.Xpp) * self.Iq) / self.T0pp) * self.base_mva / self.Sbase
            # k_Edpp = h * f4

            # Mechanical equation
            Tm = self.calc_tmech(self.slip)
            f5 = (Tm - self.Te) / (2 * self.H)
            # k_s = h * f5
        
        else:
            f1 = np.zeros_like(Edp)
            f2 = np.zeros_like(Edp)
            f3 = np.zeros_like(Edp)
            f4 = np.zeros_like(Edp)
            f5 = np.zeros_like(Edp)

        return f1, f2, f3, f4, f5

    def check_diffs(self):
        """
        Check if differential equations are zero (on initialisation)
        """

        # State variables
        # Eqp_0 = self.Eqp
        # Edp_0 = self.Edp
        # self.slip = self.slip
        # 
        # Id = self.Id
        # Iq = self.Iq
        # Te = self.Te
        # 
        # Rs = self.Rs
        # X0 = self.X0
        # T0p = self.T0p
        # Xp = self.Xp

        dEdp = self.omega_n * self.slip * self.Eqp - (self.Edp + (self.X0 - self.Xp) * self.Iq) / self.T0p
        dEqp = -self.omega_n * self.slip * self.Edp - (self.Eqp - (self.X0 - self.Xp) * self.Id) / self.T0p
        ds = self.calc_tmech(1) - self.Te

        if np.round(dEdp, 6).all() != 0 or np.round(dEqp, 6).all() != 0 or np.round(ds, 6).all() != 0:
            print('Warning: differential equations not zero on initialisation...')
            print('dEdp = ' + str(dEdp) + ', dEqp = ' + str(dEqp) + ', ds = ' + str(ds))


def dynamic_simulation(n, Vbus, Sbus, Ybus, Sbase, fBase, t_sim, h, dynamic_devices=list(), bus_indices=list(),
                       callback=None):
    """
    Dynamic transient simulation of a power system
    Args:
        n: number of nodes
        Vbus:
        Ybus:
        Sbase:
        fBase: base frequency i.e. 50Hz
        t_sim:
        h:
        dynamic_devices: objects of each machine
        bus_indices:

    Returns:

    """
    max_err = 1e-3
    max_iter = 20

    # compose dynamic controllers
    '''
    class DynamicModels(Enum):
    NoModel = 0,
    SM4 = 1,
    SM6b = 2,
    VSC = 3,
    EG = 4,
    SAM = 5,
    DAM = 6  
    '''
    sm4_idx = list()
    sm6b_idx = list()
    vsc_idx = list()
    eg_idx = list()
    sam_idx = list()
    dam_idx = list()

    sm4_bus_idx = list()
    sm6b_bus_idx = list()
    vsc_bus_idx = list()
    eg_bus_idx = list()
    sam_bus_idx = list()
    dam_bus_idx = list()

    n_obj = len(dynamic_devices)
    H = np.zeros(n_obj)
    a = np.zeros(n_obj)
    Xm = np.zeros(n_obj)
    Ra = np.zeros(n_obj)
    Rs = np.zeros(n_obj)
    Xs = np.zeros(n_obj)
    Xd = np.zeros(n_obj)
    Xa = np.zeros(n_obj)
    Xdp = np.zeros(n_obj)
    Xdpp = np.zeros(n_obj)
    Xq = np.zeros(n_obj)
    Xqp = np.zeros(n_obj)
    Xqpp = np.zeros(n_obj)
    Td0p = np.zeros(n_obj)
    Tq0p = np.zeros(n_obj)
    Td0pp = np.zeros(n_obj)
    Tq0pp = np.zeros(n_obj)
    base_mva = np.zeros(n_obj)

    Rr = np.zeros(n_obj)
    Rr2 = np.zeros(n_obj)
    Xr = np.zeros(n_obj)
    Xr2 = np.zeros(n_obj)
    speed_volt = np.zeros(n_obj, dtype=bool)

    machine_types = [None] * n_obj

    # extract the parameters from the objects into the arrays
    for k, machine in enumerate(dynamic_devices):

        # store the machine model in a list for later
        machine_types[k] = machine.machine_model

        # store the machine data into the representing vector
        if machine.machine_model == DynamicModels.NoModel:  # no model
            pass

        elif machine.machine_model == DynamicModels.SynchronousGeneratorOrder4:  # fourth order synchronous machine

            H[k] = dynamic_devices[k].H
            Ra[k] = dynamic_devices[k].Ra
            Xd[k] = dynamic_devices[k].Xd
            Xa[k] = dynamic_devices[k].Xa
            Xdp[k] = dynamic_devices[k].Xdp
            Xdpp[k] = dynamic_devices[k].Xdpp
            Xq[k] = dynamic_devices[k].Xq
            Xqp[k] = dynamic_devices[k].Xqp
            Xqpp[k] = dynamic_devices[k].Xqpp
            Td0p[k] = dynamic_devices[k].Td0p
            Tq0p[k] = dynamic_devices[k].Tq0p
            base_mva[k] = dynamic_devices[k].Snom
            speed_volt[k] = dynamic_devices[k].speed_volt
            sm4_idx.append(k)
            sm4_bus_idx.append(bus_indices[k])

        elif machine.machine_model == DynamicModels.SynchronousGeneratorOrder6:  # sixth order synchronous machine

            H[k] = dynamic_devices[k].H
            Ra[k] = dynamic_devices[k].Ra
            Xd[k] = dynamic_devices[k].Xd
            Xdp[k] = dynamic_devices[k].Xdp
            Xdpp[k] = dynamic_devices[k].Xdpp
            Xq[k] = dynamic_devices[k].Xq
            Xqp[k] = dynamic_devices[k].Xqp
            Xqpp[k] = dynamic_devices[k].Xqpp
            Td0p[k] = dynamic_devices[k].Td0p
            Tq0p[k] = dynamic_devices[k].Tq0p
            Td0pp[k] = dynamic_devices[k].Td0pp
            Tq0pp[k] = dynamic_devices[k].Tq0pp
            base_mva[k] = dynamic_devices[k].Snom
            speed_volt[k] = dynamic_devices[k].speed_volt
            sm6b_idx.append(k)
            sm6b_bus_idx.append(bus_indices[k])

        elif machine.machine_model == DynamicModels.VoltageSourceConverter:  # voltage source converter

            # R1, X1, fn
            Ra[k] = dynamic_devices[k].R1
            Xd[k] = dynamic_devices[k].X1
            vsc_idx.append(k)
            vsc_bus_idx.append(bus_indices[k])

        elif machine.machine_model == DynamicModels.ExternalGrid:  # external grid
            # Xdp, H
            H[k] = dynamic_devices[k].H
            Xdp[k] = dynamic_devices[k].Xdp
            eg_idx.append(k)
            eg_bus_idx.append(bus_indices[k])

        elif machine.machine_model == DynamicModels.AsynchronousSingleCageMotor:  # single cage asynchronous motor
            # H, Rr, Xr, Rs, Xs, a, Xm, Sbase, MVA_Rating
            H[k] = dynamic_devices[k].H
            Rr[k] = dynamic_devices[k].Rr
            Xr[k] = dynamic_devices[k].Xr
            Rs[k] = dynamic_devices[k].Rs
            Xs[k] = dynamic_devices[k].Xs
            Xq[k] = dynamic_devices[k].Xq
            a[k] = dynamic_devices[k].a
            Xm[k] = dynamic_devices[k].Xm
            base_mva[k] = dynamic_devices[k].MVA_Rating
            sam_idx.append(k)
            sam_bus_idx.append(bus_indices[k])

        elif machine.machine_model == DynamicModels.AsynchronousDoubleCageMotor:  # double cage asynchronous motor
            # H, Rr, Xr, Rs, Xs, a, Xm, Rr2, Xr2, MVA_Rating, Sbase
            H[k] = dynamic_devices[k].H
            Rr[k] = dynamic_devices[k].Rr
            Xr[k] = dynamic_devices[k].Xr
            Rs[k] = dynamic_devices[k].Rs
            Xs[k] = dynamic_devices[k].Xs
            Rr2[k] = dynamic_devices[k].Rr2
            Xr2[k] = dynamic_devices[k].Xr2
            a[k] = dynamic_devices[k].a
            Xm[k] = dynamic_devices[k].Xm
            base_mva[k] = dynamic_devices[k].MVA_Rating
            dam_idx.append(k)
            dam_bus_idx.append(bus_indices[k])

    # create the controllers
    sm4 = SynchronousMachineOrder4(H=H[sm4_idx],
                                   Ra=Ra[sm4_idx],
                                   Xd=Xd[sm4_idx],
                                   Xdp=Xdp[sm4_idx],
                                   Xdpp=Xdpp[sm4_idx],
                                   Xq=Xq[sm4_idx],
                                   Xqp=Xqp[sm4_idx],
                                   Xqpp=Xqp[sm4_idx],
                                   Td0p=Td0p[sm4_idx],
                                   Tq0p=Tq0p[sm4_idx],
                                   base_mva=base_mva[sm4_idx],
                                   Sbase=Sbase,
                                   bus_idx=sm4_bus_idx,
                                   fn=fBase,
                                   speed_volt=speed_volt[sm4_idx])

    sm6 = SynchronousMachineOrder6SauerPai(H=H[sm6b_idx],
                                           Ra=Ra[sm6b_idx],
                                           Xa=Xa[sm6b_idx],
                                           Xd=Xd[sm6b_idx],
                                           Xdp=Xdp[sm6b_idx],
                                           Xdpp=Xdpp[sm6b_idx],
                                           Xq=Xq[sm6b_idx],
                                           Xqp=Xqp[sm6b_idx],
                                           Xqpp=Xqpp[sm6b_idx],
                                           Td0p=Td0p[sm6b_idx],
                                           Tq0p=Tq0p[sm6b_idx],
                                           Td0pp=Td0pp[sm6b_idx],
                                           Tq0pp=Tq0pp[sm6b_idx],
                                           base_mva=base_mva[sm6b_idx],
                                           Sbase=Sbase,
                                           bus_idx=sm6b_bus_idx,
                                           fn=fBase,
                                           speed_volt=speed_volt[sm6b_idx])

    vsc = VoltageSourceConverterAverage(Rl=Ra[vsc_idx], Xl=Xd[vsc_idx], fn=fBase, bus_idx=vsc_bus_idx)

    exg = ExternalGrid(Xdp=Xdp[eg_idx], H=H[eg_idx], fn=fBase, bus_idx=eg_bus_idx)

    sam = SingleCageAsynchronousMotor(H=H[sam_idx],
                                      Rr=Rr[sam_idx],
                                      Xr=Xr[sam_idx],
                                      Rs=Rs[sam_idx],
                                      Xs=Xs[sam_idx],
                                      a=a[sam_idx],
                                      Xm=Xm[sam_idx],
                                      MVA_Rating=base_mva[sam_idx],
                                      Sbase=Sbase,
                                      bus_idx=sam_bus_idx,
                                      fn=fBase)

    dam = DoubleCageAsynchronousMotor(H=H[dam_idx],
                                      Rr=Rr[dam_idx],
                                      Xr=Xr[dam_idx],
                                      Rs=Rs[dam_idx],
                                      Xs=Xs[dam_idx],
                                      a=a[dam_idx],
                                      Xm=Xm[dam_idx],
                                      Rr2=Rr2[dam_idx],
                                      Xr2=Xr2[dam_idx],
                                      MVA_Rating=base_mva[dam_idx],
                                      Sbase=Sbase,
                                      bus_idx=dam_bus_idx,
                                      fn=fBase)

    # modify Ybus to add the admittances

    Y_shunt = np.zeros(n, dtype=complex)
    load_idx = np.where(Sbus > 0)[0]
    Y_shunt[load_idx] = Sbus[load_idx] / np.power(Vbus[load_idx], 2)  # add loads as admittances

    Y_shunt[sm4.bus_idx] += sm4.get_yg()
    Y_shunt[sm6.bus_idx] += sm6.get_yg()
    Y_shunt[vsc.bus_idx] += vsc.get_yg()
    Y_shunt[exg.bus_idx] += exg.get_yg()
    Y_shunt[sam.bus_idx] += sam.get_yg()
    Y_shunt[dam.bus_idx] += dam.get_yg()

    ib = range(n)
    Ydiag = sparse((Y_shunt, (ib, ib)))

    # factorize the impedance matrix
    Zbus = splu(Ybus + Ydiag)

    # copy the initial voltage
    V = Vbus.copy()
    I = np.zeros(n, dtype=complex)

    # initialize machines
    sm4.initialise(vt0=Vbus[sm4_bus_idx], S0=Sbus[sm4_bus_idx])
    sm6.initialise(vt0=Vbus[sm6b_bus_idx], S0=Sbus[sm6b_bus_idx])
    vsc.initialise(vt0=Vbus[vsc_bus_idx], S0=Sbus[vsc_bus_idx])
    exg.initialise(vt0=Vbus[eg_bus_idx], S0=Sbus[eg_bus_idx])
    sam.initialise(vt0=Vbus[sam_bus_idx], S0=Sbus[sam_bus_idx])
    dam.initialise(vt0=Vbus[dam_bus_idx], S0=Sbus[dam_bus_idx])

    voltages = list()
    omegas = list()

    # iterate
    t = 0.0
    time = list()
    while t < t_sim:

        sm4.solve(h)

        # compute machine currents

        sm4.calc_currents(V, I)

        # solve voltages
        V = Zbus.solve(I)

        voltages.append(V)
        omegas.append(sm4.omega)
        time.append(t)

        t += h

        if callback is not None:
            progress = t / t_sim * 100
            txt = 'Running transient stability t:' + str(t)
            callback(txt, progress)

    res = TransientStabilityResults()
    res.voltage = np.array(voltages)
    res.omegas = np.array(omegas)
    res.time = np.array(time)

    return res



"""
Script to update correctly the main GUI (.py) file from the Qt design (.ui) file
"""
import os
import sys
from subprocess import call


def correct_file_imports(filename):
    """
    Correct file with qtpy agnostic imports
    :param filename: file name
    :return: Nothing
    """
    with open(filename, 'r') as file:
        file_data = file.read()

    # Replace the target string
    file_data = file_data.replace('import icons_rc', 'from .icons_rc import *')
    file_data = file_data.replace('PySide2', 'qtpy')
    file_data = file_data.replace('PySide6', 'qtpy')
    file_data = file_data.replace('PyQt5', 'qtpy')
    file_data = file_data.replace('PyQt6', 'qtpy')

    # Write the file out again
    with open(filename, 'w') as file:
        file.write(file_data)


def convert_resource_file(source, rcc_cmd='pyside2-rcc'):

    folder = os.path.dirname(sys.executable)
    f1 = folder.split(os.sep)[-1]

    if f1 == 'bin':
        fbase = folder
    else:
        if 'script' in folder.lower():
            fbase = folder
        else:
            fbase = os.path.join(folder, 'Script')

    # get the target fil name
    target = source.replace('.qrc', '_rc.py')

    # define the possible commands
    possible_cmds = [os.path.join(fbase, rcc_cmd),
                     os.path.join(fbase, rcc_cmd + '.exe'),
                     rcc_cmd]

    for cmd in possible_cmds:

        try:
            call([sys.executable, os.path.join(fbase, cmd), source, '-o', target])
            correct_file_imports(target)
            print(rcc_cmd, ' (py) ok')
            return True
        except:
            print('Failed with', rcc_cmd)


def convert_ui_file(source, uic_cmd='pyside2-uic'):
    """
    Convert UI file to .py with qtpy agnostic imports
    :param source:
    :param uic_cmd:
    :return:
    """
    folder = os.path.dirname(sys.executable)
    f1 = folder.split(os.sep)[-1]

    if f1 == 'bin':
        fbase = folder
    else:
        if 'script' in folder.lower():
            fbase = folder
        else:
            fbase = os.path.join(folder, 'Script')

    # get the target fil name
    target = source.replace('.ui', '.py')

    # define the possible commands
    possible_cmds = [os.path.join(fbase, uic_cmd),
                     os.path.join(fbase, uic_cmd + '.exe'),
                     uic_cmd]

    for cmd in possible_cmds:

        try:
            call([sys.executable, os.path.join(fbase, cmd), source, '-o', target])
            correct_file_imports(filename=target)
            print(uic_cmd, ' (py) ok')
            return True

        except:
            print('Failed with', uic_cmd)





# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from typing import Dict
import pandas as pd
import numpy as np
from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
import GridCal.Engine.Devices as dev
from GridCal.Engine.Devices import DeviceType


def get_objects_dictionary():
    """
    creates a dictionary with the types and the circuit objects
    :return: Dictionary instance
    """
    object_types = {'area': dev.Area(),

                    'zone': dev.Zone(),

                    'substation': dev.Substation(),

                    'country': dev.Country(),

                    'bus': dev.Bus(),

                    'load': dev.Load(),

                    'static_generator': dev.StaticGenerator(),

                    'battery': dev.Battery(),

                    'generator': dev.Generator(),

                    'shunt': dev.Shunt(),

                    'wires': dev.Wire(),

                    'overhead_line_types': dev.Tower(),

                    'underground_cable_types': dev.UndergroundLineType(),

                    'sequence_line_types': dev.SequenceLineType(),

                    'transformer_types': dev.TransformerType(),

                    'branch': dev.Branch(),

                    'transformer2w': dev.Transformer2W(),

                    'line': dev.Line(),

                    'dc_line': dev.DcLine(None, None),

                    'hvdc': dev.HvdcLine(),

                    'vsc': dev.VSC(None, None),

                    'upfc': dev.UPFC(None, None),
                    }

    return object_types


def create_data_frames(circuit: MultiCircuit):
    """
    Pack the circuit information into tables (DataFrames)
    :param circuit: MultiCircuit instance
    :return: dictionary of DataFrames
    """
    dfs = dict()

    # configuration ################################################################################################
    obj = list()
    obj.append(['BaseMVA', circuit.Sbase])
    obj.append(['Version', 4])
    obj.append(['Name', str(circuit.name)])
    obj.append(['Comments', str(circuit.comments)])

    # increase the model version
    circuit.model_version += 1

    obj.append(['ModelVersion', str(circuit.model_version)])
    obj.append(['UserName', str(circuit.user_name)])
    obj.append(['program', 'GridCal'])

    dfs['config'] = pd.DataFrame(data=obj, columns=['Property', 'Value'], dtype=str)

    # get the master time profile
    T = circuit.time_profile

    ########################################################################################################
    # retrieve buses information that is necessary
    ########################################################################################################
    # names_count = dict()
    if len(circuit.buses) > 0:
        for elm in circuit.buses:

            # check name: if the name is repeated, change it so that it is not
            # if elm.name in names_count.keys():
            #     names_count[elm.name] += 1
            #     elm.name = elm.name + '_' + str(names_count[elm.name])
            # else:
            #     names_count[elm.name] = 1

            elm.ensure_area_objects(circuit)
            elm.ensure_profiles_exist(T)
            elm.retrieve_graphic_position()

    ########################################################################################################
    # declare objects to iterate  name: [sample object, list of objects, headers]
    ########################################################################################################
    object_types = get_objects_dictionary()

    # forget abut the Branch object when saving, now we have lines and transformers separated in their own lists
    del object_types['branch']

    ########################################################################################################
    # generic object iteration
    ########################################################################################################
    for object_type_name in object_types.keys():

        object_sample = object_types[object_type_name]

        headers = object_sample.editable_headers.keys()

        lists_of_objects = circuit.get_elements_by_type(object_sample.device_type)

        obj = list()
        profiles = dict()
        object_idtags = list()
        if len(lists_of_objects) > 0:

            for k, elm in enumerate(lists_of_objects):

                # get the object normal information
                obj.append(elm.get_save_data())
                object_idtags.append(elm.idtag)

                if T is not None:
                    nt = len(T)
                    if nt > 0:

                        elm.ensure_profiles_exist(T)

                        for profile_property in object_sample.properties_with_profile.values():

                            # get the array
                            arr = getattr(elm, profile_property)

                            if profile_property not in profiles.keys():
                                # create the profile
                                profiles[profile_property] = np.zeros((nt, len(lists_of_objects)), dtype=arr.dtype)

                            # copy the object profile to the array of profiles
                            profiles[profile_property][:, k] = arr

            # convert the objects' list to an array
            dta = np.array(obj)
        else:
            # declare an empty array
            dta = np.zeros((0, len(headers)))

        # declare the DataFrames for the normal data
        dfs[object_type_name] = pd.DataFrame(data=dta, columns=headers)

        # create the profiles' DataFrames
        for prop, data in profiles.items():
            dfs[object_type_name + '_' + prop] = pd.DataFrame(data=data, columns=object_idtags, index=T)

    # towers and wires -------------------------------------------------------------------------------------------------
    # because each tower contains a reference to a number of wires, these relations need to be stored as well
    associations = list()
    for tower in circuit.overhead_line_types:
        for wire in tower.wires_in_tower:
            associations.append([tower.name, wire.name, wire.xpos, wire.ypos, wire.phase])

    dfs['tower_wires'] = pd.DataFrame(data=associations, columns=['tower_name', 'wire_name', 'xpos', 'ypos', 'phase'])

    # Time -------------------------------------------------------------------------------------------------------------

    if circuit.time_profile is not None:
        if isinstance(circuit.time_profile, pd.DatetimeIndex):
            time_df = pd.DataFrame(data=circuit.time_profile.values, columns=['Time'])
        else:
            time_df = pd.DataFrame(data=circuit.time_profile, columns=['Time'])
        dfs['time'] = time_df

    return dfs


def data_frames_to_circuit(data: Dict):
    """
    Interpret data dictionary
    :param data: dictionary of data frames
    :return: MultiCircuit instance
    """
    # create circuit
    circuit = MultiCircuit()

    if 'name' in data.keys():
        circuit.name = str(data['name'])
        if circuit.name == 'nan':
            circuit.name = ''

    # set the base magnitudes
    if 'baseMVA' in data.keys():
        circuit.Sbase = data['baseMVA']

    # Set comments
    if 'Comments' in data.keys():
        circuit.comments = str(data['Comments'])
        if circuit.comments == 'nan':
            circuit.comments = ''

    if 'ModelVersion' in data.keys():
        circuit.model_version = int(data['ModelVersion'])

    if 'UserName' in data.keys():
        circuit.user_name = data['UserName']

    # dictionary of objects to iterate
    object_types = get_objects_dictionary()

    circuit.logger = Logger()

    # time profile -----------------------------------------------------------------------------------------------------
    if 'time' in data.keys():
        time_df = data['time']
        circuit.time_profile = pd.to_datetime(time_df.values[:, 0], dayfirst=True)
    else:
        circuit.time_profile = None

    # dictionary of dictionaries by element type
    # elements_dict[DataType][element_name] = actual object
    elements_dict = dict()
    elements_dict_by_name = dict()

    # ------------------------------------------------------------------------------------------------------------------
    # for each element type...
    for key, template_elm in object_types.items():

        if key in data.keys():

            # get the DataFrame
            df = data[key]

            # create the objects ...
            devices = list()
            devices_dict = dict()
            if 'idtag' in df.columns.values:
                for i in range(df.shape[0]):

                    elm = type(template_elm)()
                    idtag = df['idtag'].values[i]

                    # create the buses dictionary, this works because the bus is the first key in "object_types"
                    devices_dict[idtag] = elm

                    # add the device to the elements
                    devices.append(elm)
            else:
                for i in range(df.shape[0]):

                    elm = type(template_elm)()
                    idtag = df['name'].values[i]

                    # create the buses dictionary, this works because the bus is the first key in "object_types"
                    devices_dict[idtag] = elm

                    # add the device to the elements
                    devices.append(elm)

            elements_dict[template_elm.device_type] = devices_dict

            # fill in the objects
            if df.shape[0] > 0:

                # for each property ...
                for object_property_name, gc_prop in template_elm.editable_headers.items():

                    # if the object property exists in the data file, set all the object's property
                    if object_property_name in df.columns.values:

                        # get the type converter
                        dtype = gc_prop.tpe

                        # for each object, set the property
                        for i in range(df.shape[0]):

                            # convert and assign the data
                            if dtype is None:
                                val = df[object_property_name].values[i]
                                setattr(devices[i], object_property_name, val)

                            elif dtype in [DeviceType.SubstationDevice,
                                           DeviceType.AreaDevice,
                                           DeviceType.ZoneDevice,
                                           DeviceType.CountryDevice]:

                                """
                                This piece is to assign the objects matching the Area, Substation, Zone and Country
                                The cases may be:
                                a) there is a matching id tag -> ok, assign it
                                b) the value is a string -> create the relevant object, 
                                                            make sure it is not repeated by name
                                                            inset the object in its matching object dictionary
                                """

                                # search for the Substation, Area, Zone or Country matching object and assign the object
                                # this is the stored string (either idtag or name...)
                                val = str(df[object_property_name].values[i])

                                if dtype not in elements_dict.keys():
                                    elements_dict[dtype] = dict()

                                if dtype not in elements_dict_by_name.keys():
                                    elements_dict_by_name[dtype] = dict()

                                if val in elements_dict[dtype].keys():
                                    # the grouping exists as object, use it
                                    grouping = elements_dict[dtype][val]
                                else:
                                    # create the grouping

                                    if val in elements_dict_by_name[dtype].keys():
                                        grouping = elements_dict_by_name[dtype][val]

                                    else:
                                        grouping = type(object_types[dtype.value.lower()])(name=val)
                                        elements_dict[dtype][grouping.idtag] = grouping

                                        # store also by name
                                        elements_dict_by_name[dtype][grouping.name] = grouping

                                # set the object
                                setattr(devices[i], object_property_name, grouping)

                            elif dtype == DeviceType.BusDevice:

                                # check if the bus is in the dictionary...
                                if df[object_property_name].values[i] in elements_dict[DeviceType.BusDevice].keys():

                                    parent_bus: dev.Bus = elements_dict[DeviceType.BusDevice][df[object_property_name].values[i]]
                                    setattr(devices[i], object_property_name, parent_bus)

                                    # add the device to the bus
                                    if template_elm.device_type in [DeviceType.LoadDevice,
                                                                    DeviceType.GeneratorDevice,
                                                                    DeviceType.BatteryDevice,
                                                                    DeviceType.StaticGeneratorDevice,
                                                                    DeviceType.ShuntDevice,
                                                                    DeviceType.ExternalGridDevice]:

                                        parent_bus.add_device(devices[i])

                                else:
                                    circuit.logger.add_error('Bus not found', str(df[object_property_name].values[i]))

                            elif dtype in [DeviceType.TransformerTypeDevice,  # template types mostly
                                           DeviceType.SequenceLineDevice,
                                           DeviceType.TowerDevice]:

                                if df[object_property_name].values[i] in elements_dict[dtype].keys():

                                    # get the actual template and set it
                                    val = elements_dict[dtype][df[object_property_name].values[i]]
                                    setattr(devices[i], object_property_name, val)

                                else:
                                    circuit.logger.add_error(dtype.value + ' type not found',
                                                             str(df[object_property_name].values[i]))

                            elif dtype == bool:
                                # regular types (int, str, float, etc...)
                                val = df[object_property_name].values[i]
                                if val == 'False':
                                    setattr(devices[i], object_property_name, False)
                                elif val == 'True':
                                    setattr(devices[i], object_property_name, True)
                                else:
                                    setattr(devices[i], object_property_name, bool(val))

                            else:
                                # regular types (int, str, float, etc...)
                                try:
                                    val = dtype(df[object_property_name].values[i])
                                    setattr(devices[i], object_property_name, val)
                                except ValueError:
                                    circuit.logger.add_error('type error', devices[i].name, df[object_property_name].values[i])

                        # search the profiles in the data and assign them
                        if object_property_name in template_elm.properties_with_profile.keys():

                            # get the profile property
                            prop_prof = template_elm.properties_with_profile[object_property_name]

                            # build the profile property file-name to get it from the data
                            profile_name = key + '_' + prop_prof

                            if profile_name in data.keys():

                                # get the profile DataFrame
                                dfp = data[profile_name]

                                # for each object, set the profile
                                for i in range(dfp.shape[1]):
                                    profile = dfp.values[:, i]
                                    setattr(devices[i], prop_prof, profile.astype(dtype))

                            else:
                                circuit.logger.add_error('Profile was not found in the data', object_property_name)

                    else:
                        circuit.logger.add_error(object_property_name + ' of object type ' + str(template_elm.device_type) +
                                                 ' not found in the input data')
            else:
                # no objects of this type
                pass

            # ensure profiles existence
            if circuit.time_profile is not None:
                for i in range(df.shape[0]):
                    devices[i].ensure_profiles_exist(circuit.time_profile)

            # add the objects to the circuit (buses, branches ot template types)
            if template_elm.device_type == DeviceType.BusDevice:
                circuit.buses = devices

            # elif template_elm.device_type == DeviceType.SubstationDevice:
            #     circuit.substations = devices
            #
            # elif template_elm.device_type == DeviceType.AreaDevice:
            #     circuit.areas = devices
            #
            # elif template_elm.device_type == DeviceType.ZoneDevice:
            #     circuit.zones = devices
            #
            # elif template_elm.device_type == DeviceType.CountryDevice:
            #     circuit.countries = devices

            elif template_elm.device_type == DeviceType.BranchDevice:
                for d in devices:
                    circuit.add_branch(d)  # each branch needs to be converted accordingly

            elif template_elm.device_type == DeviceType.LineDevice:
                circuit.lines = devices

            elif template_elm.device_type == DeviceType.DCLineDevice:
                circuit.dc_lines = devices

            elif template_elm.device_type == DeviceType.Transformer2WDevice:
                circuit.transformers2w = devices

            elif template_elm.device_type == DeviceType.HVDCLineDevice:
                circuit.hvdc_lines = devices

            elif template_elm.device_type == DeviceType.UpfcDevice:
                circuit.upfc_devices = devices

            elif template_elm.device_type == DeviceType.VscDevice:
                for dev in devices:
                    dev.correct_buses_connection()
                circuit.vsc_devices = devices

            elif template_elm.device_type == DeviceType.TowerDevice:
                circuit.overhead_line_types = devices

            elif template_elm.device_type == DeviceType.TransformerTypeDevice:
                circuit.transformer_types = devices

            elif template_elm.device_type == DeviceType.UnderGroundLineDevice:
                circuit.underground_cable_types = devices

            elif template_elm.device_type == DeviceType.SequenceLineDevice:
                circuit.sequence_line_types = devices

            elif template_elm.device_type == DeviceType.WireDevice:
                circuit.wire_types = devices

        else:
            circuit.logger.add_error('The data does not contain information about the type', str(key))

    # fill in wires into towers ----------------------------------------------------------------------------------------
    if 'tower_wires' in data.keys():
        df = data['tower_wires']

        for i in range(df.shape[0]):
            tower_name = df['tower_name'].values[i]
            wire_name = df['wire_name'].values[i]

            if (tower_name in elements_dict[DeviceType.TowerDevice].keys()) and \
                    (wire_name in elements_dict[DeviceType.WireDevice].keys()):

                tower = elements_dict[DeviceType.TowerDevice][tower_name]
                wire = elements_dict[DeviceType.WireDevice][wire_name]
                xpos = df['xpos'].values[i]
                ypos = df['ypos'].values[i]
                phase = df['phase'].values[i]

                w = dev.WireInTower(wire=wire, xpos=xpos, ypos=ypos, phase=phase)
                tower.wires_in_tower.append(w)

    # Other actions ----------------------------------------------------------------------------------------------------
    circuit.logger += circuit.apply_all_branch_types()

    # Add the groups ---------------------------------------------------------------------------------------------------
    if DeviceType.SubstationDevice in elements_dict.keys():
        circuit.substations = list(elements_dict[DeviceType.SubstationDevice].values())

    if DeviceType.AreaDevice in elements_dict.keys():
        circuit.areas = list(elements_dict[DeviceType.AreaDevice].values())

    if DeviceType.ZoneDevice in elements_dict.keys():
        circuit.zones = list(elements_dict[DeviceType.ZoneDevice].values())

    if DeviceType.CountryDevice in elements_dict.keys():
        circuit.countries = list(elements_dict[DeviceType.CountryDevice].values())

    return circuit


if __name__ == '__main__':
    from GridCal.Engine.IO.file_handler import FileOpen, FileSave

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Some distribution grid.gridcal'

    main_circuit = FileOpen(fname).open()

    FileSave(main_circuit, 'file.gridcal').save()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import os
import chardet
import pandas as pd
from math import sqrt
from typing import Set, Dict, List, Tuple
from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.IO.zip_interface import get_xml_from_zip, get_xml_content
from GridCal.Engine.Core.multi_circuit import MultiCircuit
import GridCal.Engine.Devices as gcdev
import GridCal.Engine.IO.cim.cim_devices as cimdev
from GridCal.Engine.IO.cim.cim_circuit import CIMCircuit


def read_cim_files(cim_files):
    """
    Reads a list of .zip or xml into a dictionary of file name -> list of text lines
    :param cim_files: list of file names
    :return: dictionary of file name -> list of text lines
    """
    # read files and sort them in the preferred reading order
    data = dict()

    if isinstance(cim_files, list):

        for f in cim_files:
            _, file_extension = os.path.splitext(f)
            name = os.path.basename(f)

            if file_extension == '.xml':
                file_ptr = open(f, 'rb')
                data[name] = get_xml_content(file_ptr)
                file_ptr.close()
            elif file_extension == '.zip':
                # read the content of a zip file
                d = get_xml_from_zip(file_name_zip=f)
                for key, value in d.items():
                    data[key] = value
    else:
        name, file_extension = os.path.splitext(cim_files)

        if file_extension == '.xml':
            file_ptr = open(cim_files, 'rb')
            data[name] = get_xml_content(file_ptr)
            file_ptr.close()

        elif file_extension == '.zip':
            # read the content of a zip file
            d = get_xml_from_zip(file_name_zip=cim_files)
            for key, value in d.items():
                data[key] = value

    return data


def sort_cim_files(file_names):
    """
    Sorts the CIM files in the preferred reading order
    :param file_names: lis of file names
    :return: sorted list of file names
    """
    # sort the files
    lst = list()
    nn = len(file_names)
    for i in range(nn - 1, -1, -1):
        f = file_names[i]
        if 'TP' in f or 'TPDB' in f:
            lst.append(file_names.pop(i))

    nn = len(file_names)
    for i in range(nn - 1, -1, -1):
        f = file_names[i]
        if 'EQ' in f or 'EQBD' in f:
            lst.append(file_names.pop(i))

    lst2 = lst + file_names

    return lst2


def get_elements(d: Dict, keys):

    elm = list()

    for k in keys:
        try:
            lst = d[k]
            elm += lst
        except KeyError:
            pass

    return elm


def any_in_dict(d: Dict, keys):

    found = False

    for k in keys:
        try:
            lst = d[k]
            found = True
        except KeyError:
            pass

    return found


def try_buses(b1, b2, bus_duct):

    try:
        B1 = bus_duct[b1]
    except KeyError:
        B1 = None

    try:
        B2 = bus_duct[b2]
    except KeyError:
        B2 = None

    return B1, B2


def try_bus(b1, bus_duct):
    try:
        B1 = bus_duct[b1]
    except KeyError:
        B1 = None

    return B1


class CIMExport:

    def __init__(self, circuit: MultiCircuit):

        self.circuit = circuit

        self.logger = Logger()

    def save(self, file_name):
        """
        Save XML CIM version of a grid
        Args:
            file_name: file path
        """

        # open CIM file for writing
        text_file = open(file_name, "w")

        # header
        text_file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        text_file.write('<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:cim="http://iec.ch/TC57/2009/CIM-schema-cim14#">\n')

        # Model
        model = cimdev.GeneralContainer(rfid=self.circuit.idtag, tpe='Model')
        model.properties['name'] = self.circuit.name
        model.properties['version'] = 1
        model.properties['description'] = self.circuit.comments
        text_file.write(model.get_xml(1))

        bus_id_dict = dict()
        base_voltages = set()
        base_voltages_dict = dict()

        # dictionary of substation given a bus
        substation_bus = dict()

        # buses sweep to gather previous data (base voltages, etc..)
        for i, bus in enumerate(self.circuit.buses):

            Vnom = bus.Vnom

            # add the nominal voltage to the set of bus_voltages
            base_voltages.add(Vnom)

            # if the substation was not accounted for, create the list of voltage levels
            if bus.substation not in substation_bus.keys():
                substation_bus[bus.substation] = dict()

            if Vnom not in substation_bus[bus.substation].keys():
                substation_bus[bus.substation][Vnom] = list()

            # add bus to the categorization
            substation_bus[bus.substation][Vnom].append(bus)

        # generate Base voltages
        for V in base_voltages:

            conn_node_id = 'Base_voltage_' + str(V).replace('.', '_')

            base_voltages_dict[V] = conn_node_id

            model = cimdev.GeneralContainer(rfid=conn_node_id, tpe='BaseVoltage')
            model.properties['name'] = conn_node_id
            model.properties['nominalVoltage'] = V
            text_file.write(model.get_xml(1))

        # generate voltage levels, substations and buses and their objects
        substation_idx = 0
        voltage_level_idx = 0
        bus_idx = 0
        terminal_resources = ['TopologicalNode', 'ConductingEquipment']

        for substation in substation_bus.keys():

            substation_id = substation.idtag
            substation_idx += 1

            model = cimdev.GeneralContainer(rfid=substation_id, tpe='Substation', resources=['Location', 'SubGeographicalRegion'])
            model.properties['name'] = substation
            model.properties['aliasName'] = substation
            model.properties['PSRType'] = ''
            model.properties['Location'] = ''
            model.properties['SubGeographicalRegion'] = ''
            text_file.write(model.get_xml(1))

            for voltage_level in substation_bus[substation].keys():

                voltage_level_id = 'VoltageLevel_' + str(voltage_level).replace('.', '_') + '_' + str(voltage_level_idx)
                voltage_level_idx += 1

                base_voltage = base_voltages_dict[voltage_level]

                model = cimdev.GeneralContainer(rfid=voltage_level_id,
                                                tpe='VoltageLevel',
                                                resources=['BaseVoltage', 'Substation'])
                model.properties['name'] = substation
                model.properties['aliasName'] = substation
                model.properties['BaseVoltage'] = base_voltage
                model.properties['Substation'] = substation_id
                text_file.write(model.get_xml(1))

                # buses sweep to actually generate XML
                for bus in substation_bus[substation][voltage_level]:

                    # make id
                    conn_node_id = 'BUS_' + str(bus_idx)

                    Vnom = bus.Vnom

                    # make dictionary entry
                    bus_id_dict[bus] = conn_node_id

                    base_voltage = base_voltages_dict[Vnom]

                    if bus.Vnom <= 0.0:
                        self.logger.add_error('Zero nominal voltage', bus.name)

                    # generate model
                    model = cimdev.GeneralContainer(rfid=conn_node_id, tpe='ConnectivityNode',
                                                    resources=['BaseVoltage',
                                                               'VoltageLevel',
                                                               'ConnectivityNodeContainer'],
                                                    class_replacements={'name': 'IdentifiedObject',
                                                                        'aliasName': 'IdentifiedObject'}
                                                    )
                    model.properties['name'] = bus.name
                    model.properties['aliasName'] = bus.name
                    model.properties['BaseVoltage'] = base_voltage
                    # model.properties['VoltageLevel'] = voltage_level_id
                    model.properties['ConnectivityNodeContainer'] = voltage_level_id
                    text_file.write(model.get_xml(1))

                    for il, elm in enumerate(bus.loads):

                        id2 = elm.idtag + '_LOAD_' + str(il)
                        id3 = id2 + '_LRC'

                        model = cimdev.GeneralContainer(rfid=id2, tpe='ConformLoad',
                                                        resources=['BaseVoltage',
                                                                   'LoadResponse',
                                                                   'VoltageLevel'],
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject',
                                                                            'EquipmentContainer': 'Equipment'})
                        model.properties['name'] = elm.name
                        model.properties['aliasName'] = elm.name
                        model.properties['BaseVoltage'] = base_voltage
                        model.properties['EquipmentContainer'] = voltage_level_id
                        model.properties['LoadResponse'] = id3
                        model.properties['pfixed'] = elm.P
                        model.properties['qfixed'] = elm.Q
                        model.properties['normallyInService'] = elm.active
                        text_file.write(model.get_xml(1))

                        model = cimdev.GeneralContainer(rfid=id3, tpe='LoadResponseCharacteristic', resources=[])
                        model.properties['name'] = elm.name
                        model.properties['exponentModel'] = 'false'
                        model.properties['pConstantCurrent'] = elm.Ir
                        model.properties['pConstantImpedance'] = 1 / (elm.G + 1e-20)
                        model.properties['pConstantPower'] = elm.P
                        model.properties['pVoltageExponent'] = 0.0
                        model.properties['pFrequencyExponent'] = 0.0
                        model.properties['qConstantCurrent'] = elm.Ir
                        model.properties['qConstantImpedance'] = 1 / (elm.B + 1e-20)
                        model.properties['qConstantPower'] = elm.Q
                        model.properties['qVoltageExponent'] = 0.0
                        model.properties['qFrequencyExponent'] = 0.0
                        text_file.write(model.get_xml(1))

                        # Terminal 1 (from)
                        model = cimdev.GeneralContainer(rfid=id2 + '_T', tpe='Terminal',
                                                        resources=terminal_resources,
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject'})
                        model.properties['name'] = elm.name
                        model.properties['TopologicalNode'] = bus_id_dict[bus]
                        model.properties['ConductingEquipment'] = id2
                        model.properties['connected'] = 'true'
                        model.properties['sequenceNumber'] = '1'
                        text_file.write(model.get_xml(1))

                    for il, elm in enumerate(bus.static_generators):

                        id2 = elm.idtag + '_StatGen_' + str(il)

                        model = cimdev.GeneralContainer(rfid=id2, tpe='ConformLoad',
                                                        resources=['BaseVoltage', 'LoadResponse', 'VoltageLevel'],
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject',
                                                                            'EquipmentContainer': 'Equipment'}
                                                        )
                        model.properties['name'] = elm.name
                        model.properties['aliasName'] = elm.name
                        model.properties['BaseVoltage'] = base_voltage
                        model.properties['EquipmentContainer'] = voltage_level_id
                        model.properties['pfixed'] = -elm.P
                        model.properties['qfixed'] = -elm.Q
                        model.properties['normallyInService'] = elm.active
                        text_file.write(model.get_xml(1))

                        # Terminal 1 (from)
                        model = cimdev.GeneralContainer(rfid=id2 + '_T', tpe='Terminal',
                                                        resources=terminal_resources,
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject'}
                                                        )
                        model.properties['name'] = elm.name
                        model.properties['TopologicalNode'] = bus_id_dict[bus]
                        model.properties['ConductingEquipment'] = id2
                        model.properties['connected'] = 'true'
                        model.properties['sequenceNumber'] = '1'
                        text_file.write(model.get_xml(1))

                    for il, elm in enumerate(bus.controlled_generators):

                        id2 = elm.idtag + '_SyncGen_' + str(il)
                        id3 = id2 + '_GU'
                        id4 = id2 + '_RC'

                        model = cimdev.GeneralContainer(rfid=id2, tpe='SynchronousMachine',
                                                        resources=['BaseVoltage', 'RegulatingControl',
                                                                   'GeneratingUnit', 'VoltageLevel'],
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject',
                                                                            'EquipmentContainer': 'Equipment'}
                                                        )
                        model.properties['name'] = elm.name
                        model.properties['aliasName'] = elm.name
                        model.properties['BaseVoltage'] = base_voltage
                        model.properties['EquipmentContainer'] = voltage_level_id
                        model.properties['RegulatingControl'] = id3
                        model.properties['GeneratingUnit'] = id4
                        model.properties['maxQ'] = elm.Qmax
                        model.properties['minQ'] = elm.Qmin
                        model.properties['ratedS'] = elm.Snom
                        model.properties['normallyInService'] = elm.active
                        text_file.write(model.get_xml(1))

                        model = cimdev.GeneralContainer(rfid=id3, tpe='RegulatingControl', resources=[])
                        model.properties['name'] = elm.name
                        model.properties['targetValue'] = elm.Vset * bus.Vnom
                        text_file.write(model.get_xml(1))

                        model = cimdev.GeneralContainer(rfid=id4, tpe='GeneratingUnit', resources=[])
                        model.properties['name'] = elm.name
                        model.properties['initialP'] = elm.P
                        text_file.write(model.get_xml(1))

                        # Terminal 1 (from)
                        model = cimdev.GeneralContainer(rfid=id2 + '_T', tpe='Terminal',
                                                        resources=terminal_resources,
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject'}
                                                        )
                        model.properties['name'] = elm.name
                        model.properties['TopologicalNode'] = bus_id_dict[bus]
                        model.properties['ConductingEquipment'] = id2
                        model.properties['connected'] = 'true'
                        model.properties['sequenceNumber'] = '1'
                        text_file.write(model.get_xml(1))

                    for il, elm in enumerate(bus.shunts):

                        id2 = elm.idtag + '_Shunt_' + str(il)

                        model = cimdev.GeneralContainer(rfid=id2, tpe='ShuntCompensator',
                                                        resources=['BaseVoltage', 'VoltageLevel'],
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject',
                                                                            'EquipmentContainer': 'Equipment'}
                                                        )
                        model.properties['name'] = elm.name
                        model.properties['aliasName'] = elm.name
                        model.properties['BaseVoltage'] = base_voltage
                        model.properties['EquipmentContainer'] = voltage_level_id
                        model.properties['gPerSection'] = elm.G
                        model.properties['bPerSection'] = elm.B
                        model.properties['g0PerSection'] = 0.0
                        model.properties['b0PerSection'] = 0.0
                        model.properties['normallyInService'] = elm.active
                        text_file.write(model.get_xml(1))

                        # Terminal 1 (from)
                        model = cimdev.GeneralContainer(rfid=id2 + '_T', tpe='Terminal',
                                                        resources=terminal_resources,
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject'}
                                                        )
                        model.properties['name'] = elm.name
                        model.properties['TopologicalNode'] = bus_id_dict[bus]
                        model.properties['ConductingEquipment'] = id2
                        model.properties['connected'] = 'true'
                        model.properties['sequenceNumber'] = '1'
                        text_file.write(model.get_xml(1))

                    if bus.is_slack:
                        equivalent_network_id = conn_node_id + '_EqNetwork'

                        model = cimdev.GeneralContainer(rfid=equivalent_network_id, tpe='EquivalentNetwork',
                                                        resources=['BaseVoltage', 'VoltageLevel'],
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject'})
                        model.properties['name'] = bus.name + '_Slack'
                        model.properties['aliasName'] = bus.name + '_Slack'
                        model.properties['BaseVoltage'] = base_voltage
                        model.properties['VoltageLevel'] = voltage_level_id
                        text_file.write(model.get_xml(1))

                        # Terminal 1 (from)
                        model = cimdev.GeneralContainer(rfid=equivalent_network_id + '_T', tpe='Terminal',
                                                        resources=terminal_resources,
                                                        class_replacements={'name': 'IdentifiedObject',
                                                                            'aliasName': 'IdentifiedObject'}
                                                        )
                        model.properties['name'] = equivalent_network_id + '_T'
                        model.properties['TopologicalNode'] = bus_id_dict[bus]
                        model.properties['ConductingEquipment'] = equivalent_network_id
                        model.properties['connected'] = 'true'
                        model.properties['sequenceNumber'] = '1'
                        text_file.write(model.get_xml(1))

                    # increment the bus index
                    bus_idx += 1

        # Branches
        winding_resources = ['connectionType', 'windingType', 'PowerTransformer']
        tap_changer_resources = ['TransformerWinding']
        for i, branch in enumerate(self.circuit.transformers2w):

            model = cimdev.GeneralContainer(rfid=branch.idtag, tpe='PowerTransformer',
                                            resources=[],
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject',
                                                                'EquipmentContainer': 'Equipment'}
                                            )
            model.properties['name'] = branch.name
            model.properties['aliasName'] = branch.name
            text_file.write(model.get_xml(1))

            #  warnings
            if branch.rate <= 0.0:
                self.logger.add_error('The rate is 0', branch.name)
                raise Exception(branch.name + ": The rate is 0, this will cause a problem when loading.")

            if branch.bus_from.Vnom <= 0.0:
                self.logger.add_error('Vfrom is zero', branch.name)
                raise Exception(branch.name + ": The voltage at the from side, this will cause a problem when loading.")

            if branch.bus_to.Vnom <= 0.0:
                self.logger.add_error('Vto is zero', branch.name)
                raise Exception(branch.name + ": The voltage at the to side, this will cause a problem when loading.")

            # W1 (from)
            winding_power_rate = branch.rate / 2
            Zbase = (branch.bus_from.Vnom ** 2) / winding_power_rate
            Ybase = 1 / Zbase
            model = cimdev.GeneralContainer(rfid=branch.idtag + "_W1", tpe='PowerTransformerEnd',
                                            resources=winding_resources,
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject',
                                                                'BaseVoltage': 'ConductingEquipment'}
                                            )
            model.properties['name'] = branch.name
            model.properties['PowerTransformer'] = branch.idtag
            model.properties['BaseVoltage'] = base_voltages_dict[branch.bus_from.Vnom]
            model.properties['r'] = branch.R / 2 * Zbase
            model.properties['x'] = branch.X / 2 * Zbase
            model.properties['g'] = branch.G / 2 * Ybase
            model.properties['b'] = branch.B / 2 * Ybase
            model.properties['r0'] = 0.0
            model.properties['x0'] = 0.0
            model.properties['g0'] = 0.0
            model.properties['b0'] = 0.0
            model.properties['ratedS'] = winding_power_rate
            model.properties['ratedU'] = branch.bus_from.Vnom
            model.properties['rground'] = 0.0
            model.properties['xground'] = 0.0
            model.properties['connectionType'] = "http://iec.ch/TC57/2009/CIM-schema-cim14#WindingConnection.Y"
            model.properties['windingType'] = "http://iec.ch/TC57/2009/CIM-schema-cim14#WindingType.primary"
            text_file.write(model.get_xml(1))

            # W2 (To)
            Zbase = (branch.bus_to.Vnom ** 2) / winding_power_rate
            Ybase = 1 / Zbase
            model = cimdev.GeneralContainer(rfid=branch.idtag + "_W2", tpe='PowerTransformerEnd',
                                            resources=winding_resources,
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject',
                                                                'BaseVoltage': 'ConductingEquipment'})
            model.properties['name'] = branch.name
            model.properties['PowerTransformer'] = branch.idtag
            model.properties['BaseVoltage'] = base_voltages_dict[branch.bus_to.Vnom]
            model.properties['r'] = branch.R / 2 * Zbase
            model.properties['x'] = branch.X / 2 * Zbase
            model.properties['g'] = branch.G / 2 * Ybase
            model.properties['b'] = branch.B / 2 * Ybase
            model.properties['r0'] = 0.0
            model.properties['x0'] = 0.0
            model.properties['g0'] = 0.0
            model.properties['b0'] = 0.0
            model.properties['ratedS'] = winding_power_rate
            model.properties['ratedU'] = branch.bus_to.Vnom
            model.properties['rground'] = 0.0
            model.properties['xground'] = 0.0
            model.properties['connectionType'] = "http://iec.ch/TC57/2009/CIM-schema-cim14#WindingConnection.Y"
            model.properties['windingType'] = "http://iec.ch/TC57/2009/CIM-schema-cim14#WindingType.secondary"
            text_file.write(model.get_xml(1))

            # add tap changer at the "to" winding
            if branch.tap_module != 1.0 and branch.angle != 0.0:

                Vnom = branch.bus_to.Vnom
                SVI = (Vnom - Vnom * branch.tap_module) * 100.0 / Vnom

                model = cimdev.GeneralContainer(rfid=branch.idtag + 'Tap_2', tpe='RatioTapChanger', resources=tap_changer_resources)
                model.properties['TransformerWinding'] = branch.idtag + "_W2"
                model.properties['name'] = branch.name + 'tap changer'
                model.properties['neutralU'] = Vnom
                model.properties['stepVoltageIncrement'] = SVI
                model.properties['step'] = 0
                model.properties['lowStep'] = -1
                model.properties['highStep'] = 1
                model.properties['subsequentDelay'] = branch.angle
                text_file.write(model.get_xml(1))

            # Terminal 1 (from)
            model = cimdev.GeneralContainer(rfid=branch.idtag + '_T1', tpe='Terminal',
                                            resources=terminal_resources,
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject'}
                                            )
            model.properties['name'] = branch.name + '_T1'
            model.properties['TopologicalNode'] = bus_id_dict[branch.bus_from]
            model.properties['ConductingEquipment'] = branch.idtag
            model.properties['connected'] = 'true'
            model.properties['sequenceNumber'] = '1'
            text_file.write(model.get_xml(1))

            # Terminal 2 (to)
            model = cimdev.GeneralContainer(rfid=branch.idtag + '_T2', tpe='Terminal',
                                            resources=terminal_resources,
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject'}
                                            )
            model.properties['name'] = branch.name + '_T2'
            model.properties['TopologicalNode'] = bus_id_dict[branch.bus_to]
            model.properties['ConductingEquipment'] = branch.idtag
            model.properties['connected'] = 'true'
            model.properties['sequenceNumber'] = '1'
            text_file.write(model.get_xml(1))

        for i, branch in enumerate(self.circuit.lines):

            if branch.branch_type == gcdev.BranchType.Line or branch.branch_type == gcdev.BranchType.Branch:

                Zbase = (branch.bus_from.Vnom ** 2) / self.circuit.Sbase

                if branch.bus_from.Vnom <= 0.0:
                    Ybase = 0
                else:
                    Ybase = 1 / Zbase

                model = cimdev.GeneralContainer(rfid=branch.idtag, tpe='ACLineSegment',
                                                resources=['BaseVoltage'],
                                                class_replacements={'name': 'IdentifiedObject',
                                                                    'aliasName': 'IdentifiedObject',
                                                                    'BaseVoltage': 'ConductingEquipment',
                                                                    'value': 'CurrentLimit'}
                                                )
                model.properties['name'] = branch.name
                model.properties['aliasName'] = branch.name
                model.properties['BaseVoltage'] = base_voltages_dict[branch.bus_from.Vnom]
                model.properties['r'] = branch.R * Zbase
                model.properties['x'] = branch.X * Zbase
                model.properties['gch'] = 0
                model.properties['bch'] = branch.B * Ybase
                model.properties['r0'] = 0.0
                model.properties['x0'] = 0.0
                model.properties['g0ch'] = 0.0
                model.properties['b0ch'] = 0.0
                model.properties['length'] = 1.0
                model.properties['value'] = branch.rate / (branch.bus_from.Vnom * sqrt(3))  # kA
                text_file.write(model.get_xml(1))

            elif branch.branch_type == gcdev.BranchType.Switch:

                model = cimdev.GeneralContainer(rfid=branch.idtag, tpe='Switch', resources=['BaseVoltage'])
                model.properties['name'] = branch.name
                model.properties['aliasName'] = branch.name
                model.properties['BaseVoltage'] = base_voltages_dict[branch.bus_from.Vnom]
                model.properties['normalOpen'] = False
                model.properties['open'] = not branch.active
                text_file.write(model.get_xml(1))

            elif branch.branch_type == gcdev.BranchType.Reactance:
                self.logger.add_warning('Reactance CIM export not implemented yet, exported as a branch', branch.name)

                conn_node_id = 'Reactance_' + str(i)
                Zbase = (branch.bus_from.Vnom ** 2) / self.circuit.Sbase

                if branch.bus_from.Vnom <= 0.0:
                    Ybase = 0
                else:
                    Ybase = 1 / Zbase

                model = cimdev.GeneralContainer(rfid=branch.idtag, tpe='ACLineSegment', resources=['BaseVoltage'])
                model.properties['name'] = branch.name
                model.properties['aliasName'] = branch.name
                model.properties['BaseVoltage'] = base_voltages_dict[branch.bus_from.Vnom]
                model.properties['r'] = branch.R * Zbase
                model.properties['x'] = branch.X * Zbase
                model.properties['gch'] = branch.G * Ybase
                model.properties['bch'] = branch.B * Ybase
                model.properties['r0'] = 0.0
                model.properties['x0'] = 0.0
                model.properties['g0ch'] = 0.0
                model.properties['b0ch'] = 0.0
                model.properties['length'] = 1.0
                text_file.write(model.get_xml(1))

            # Terminal 1 (from)
            model = cimdev.GeneralContainer(rfid=branch.idtag + '_T1', tpe='Terminal',
                                            resources=terminal_resources,
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject'}
                                            )
            model.properties['name'] = bus.name + '_' + branch.name + '_T1'
            model.properties['TopologicalNode'] = bus_id_dict[branch.bus_from]
            model.properties['ConductingEquipment'] = branch.idtag
            model.properties['connected'] = 'true'
            model.properties['sequenceNumber'] = '1'
            text_file.write(model.get_xml(1))

            # Terminal 2 (to)
            model = cimdev.GeneralContainer(rfid=branch.idtag + '_T2', tpe='Terminal',
                                            resources=terminal_resources,
                                            class_replacements={'name': 'IdentifiedObject',
                                                                'aliasName': 'IdentifiedObject'}
                                            )
            model.properties['name'] = bus.name + '_' + branch.name + '_T2'
            model.properties['TopologicalNode'] = bus_id_dict[branch.bus_to]
            model.properties['ConductingEquipment'] = branch.idtag
            model.properties['connected'] = 'true'
            model.properties['sequenceNumber'] = '1'
            text_file.write(model.get_xml(1))

        # end
        text_file.write("</rdf:RDF>")

        text_file.close()


class CIMImport:

    def __init__(self, text_func=None, progress_func=None):

        self.logger = Logger()

        self.cim = CIMCircuit(logger=self.logger)

        # relations between connectivity nodes and terminals
        # node_terminal[some_node] = list of terminals
        self.node_terminal = dict()
        self.terminal_node = dict()

        self.text_func = text_func
        self.progress_func = progress_func

        self.needs_compiling = True

        self.topology = None

    def emit_text(self, val):
        if self.text_func is not None:
            self.text_func(val)

    def emit_progress(self, val):
        if self.progress_func is not None:
            self.progress_func(val)

    @staticmethod
    def parse_model(cim: CIMCircuit, circuit: MultiCircuit):
        """

        :param cim:
        :param circuit:
        :return:
        """
        if 'Model' in cim.elements_by_type.keys():
            for elm in cim.elements_by_type['Model']:
                if 'description' in elm.properties.keys():
                    circuit.comments = elm.properties['description']

                if 'name' in elm.properties.keys():
                    circuit.name = elm.properties['name']

    def parse_bus_bars(self, cim: CIMCircuit, circuit: MultiCircuit):
        """

        :param cim:
        :param circuit:
        :return:
        """

        busbar_dict = dict()

        if 'BusbarSection' in cim.elements_by_type.keys():
            for elm in cim.elements_by_type['BusbarSection']:

                obj = gcdev.Bus(name=str(elm.name),
                                idtag=elm.uuid)

                circuit.add_bus(obj)

                busbar_dict[elm] = obj
        else:
            self.logger.add_error("No BusbarSections: There is no chance to reduce the grid")

        return busbar_dict

    def parse_ac_line_segment(self, cim: CIMCircuit, circuit: MultiCircuit, busbar_dict):
        """

        :param cim:
        :param circuit:
        :param busbar_dict:
        :return:
        """

        if 'ACLineSegment' in cim.elements_by_type.keys():
            for elm in cim.elements_by_type['ACLineSegment']:

                b1, b2 = elm.get_buses()

                B1, B2 = try_buses(b1, b2, busbar_dict)

                if B1 is not None and B2 is not None:
                    R, X, G, B = elm.get_pu_values()
                    rate = elm.get_rate()

                    # create AcLineSegment (Line)
                    line = gcdev.Line(idtag=elm.uuid,
                                      bus_from=B1,
                                      bus_to=B2,
                                      name=str(elm.name),
                                      r=R,
                                      x=X,
                                      b=B,
                                      rate=rate,
                                      active=True,
                                      mttf=0,
                                      mttr=0)

                    circuit.add_line(line)
                else:
                    self.logger.add_error('Bus not found', elm.rfid)

    def parse_power_transformer(self, cim: CIMCircuit, circuit: MultiCircuit, busbar_dict):
        """

        :param cim:
        :param circuit:
        :param busbar_dict:
        :return:
        """
        if 'PowerTransformer' in cim.elements_by_type.keys():
            for elm in cim.elements_by_type['PowerTransformer']:
                b1, b2 = elm.get_buses()
                B1, B2 = try_buses(b1, b2, busbar_dict)

                if B1 is not None and B2 is not None:
                    R, X, G, B = elm.get_pu_values()
                    rate = elm.get_rate()

                    voltages = elm.get_voltages()
                    voltages.sort()

                    if len(voltages) == 2:
                        lv, hv = voltages
                    else:
                        lv = 1
                        hv = 1
                        self.logger.add_error('Could not parse transformer nominal voltages', self.name)

                    line = gcdev.Transformer2W(idtag=cimdev.rfid2uuid(elm.rfid),
                                               bus_from=B1,
                                               bus_to=B2,
                                               name=str(elm.name),
                                               r=R,
                                               x=X,
                                               g=G,
                                               b=B,
                                               rate=rate,
                                               tap=1.0,
                                               shift_angle=0,
                                               active=True,
                                               HV=hv,
                                               LV=lv)

                    circuit.add_branch(line)
                else:
                    self.logger.add_error('Bus not found', elm.rfid)

    def parse_switches(self, cim: CIMCircuit, circuit: MultiCircuit, busbar_dict):
        """

        :param cim:
        :param circuit:
        :param busbar_dict:
        :return:
        """
        EPS = 1e-20
        cim_switches = ['Switch', 'Disconnector', 'Breaker', 'LoadBreakSwitch']
        if any_in_dict(cim.elements_by_type, cim_switches):
            for elm in get_elements(cim.elements_by_type, cim_switches):
                b1, b2 = elm.get_buses()
                B1, B2 = try_buses(b1, b2, busbar_dict)

                if B1 is not None and B2 is not None:
                    state = True
                    line = gcdev.Switch(idtag=elm.uuid,
                                        bus_from=B1,
                                        bus_to=B2,
                                        name=str(elm.name),
                                        r=EPS,
                                        x=EPS,
                                        rate=EPS,
                                        active=state)

                    circuit.add_switch(line)
                else:
                    self.logger.add_error('Bus not found', elm.rfid)

    def parse_loads(self, cim: CIMCircuit, circuit: MultiCircuit, busbar_dict):
        """

        :param cim:
        :param circuit:
        :param busbar_dict:
        :return:
        """
        cim_loads = ['ConformLoad', 'EnergyConsumer', 'NonConformLoad']
        if any_in_dict(cim.elements_by_type, cim_loads):
            for elm in get_elements(cim.elements_by_type, cim_loads):

                b1 = elm.get_bus()
                B1 = try_bus(b1, busbar_dict)

                if B1 is not None:

                    p, q = elm.get_pq()

                    load = gcdev.Load(idtag=elm.uuid,
                                      name=str(elm.name),
                                      G=0,
                                      B=0,
                                      Ir=0,
                                      Ii=0,
                                      P=p if p is not None else 0,
                                      Q=q if q is not None else 0)
                    circuit.add_load(B1, load)
                else:
                    self.logger.add_error('Bus not found', elm.rfid)

    def parse_shunts(self, cim: CIMCircuit, circuit: MultiCircuit, busbar_dict):
        """

        :param cim:
        :param circuit:
        :param busbar_dict:
        :return:
        """
        if 'ShuntCompensator' in cim.elements_by_type.keys():
            for elm in cim.elements_by_type['ShuntCompensator']:
                b1 = elm.get_bus()
                B1 = try_bus(b1, busbar_dict)

                if B1 is not None:
                    g = 0
                    b = 0
                    sh = gcdev.Shunt(idtag=elm.uuid,
                                     name=str(elm.name),
                                     G=g,
                                     B=b)
                    circuit.add_shunt(B1, sh)
                else:
                    self.logger.add_error('Bus not found', elm.rfid)

    def parse_generators(self, cim: CIMCircuit, circuit: MultiCircuit, busbar_dict):
        """

        :param cim:
        :param circuit:
        :param busbar_dict:
        :return:
        """
        if 'SynchronousMachine' in cim.elements_by_type.keys():
            for elm in cim.elements_by_type['SynchronousMachine']:
                b1 = elm.get_bus()
                B1 = try_bus(b1, busbar_dict)

                if B1 is not None:

                    gen = gcdev.Generator(idtag=elm.uuid,
                                          name=str(elm.name),
                                          active_power=elm.p,
                                          voltage_module=1.0)
                    circuit.add_generator(B1, gen)

                else:
                    self.logger.add_error('Bus not found', elm.rfid)

    def load_cim_file(self, cim_files):
        """
        Load CIM file
        :param cim_files: list of CIM files (.xml)
        """

        # declare GridCal circuit
        circuit = MultiCircuit()
        EPS = 1e-16

        # declare CIM circuit to process the file(s)
        self.cim = CIMCircuit(text_func=self.text_func, progress_func=self.progress_func, logger=self.logger)

        # import the cim files' content into a dictionary
        data = read_cim_files(cim_files)

        lst2 = sort_cim_files(list(data.keys()))
        # Parse the files
        for f in lst2:
            name, file_extension = os.path.splitext(f)
            self.emit_text('Parsing xml structure of ' + name)
            self.cim.parse_xml_text(text_lines=data[f])

        # replace CIM references in the CIM objects
        self.emit_text('Looking for CIM references...')
        self.cim.find_references()

        # Parse devices into GridCal
        self.emit_text('Converting CIM to GridCal...')
        self.parse_model(self.cim, circuit)
        busbar_dict = self.parse_bus_bars(self.cim, circuit)
        self.parse_ac_line_segment(self.cim, circuit, busbar_dict)
        self.parse_ac_line_segment(self.cim, circuit, busbar_dict)
        self.parse_power_transformer(self.cim, circuit, busbar_dict)
        self.parse_switches(self.cim, circuit, busbar_dict)
        self.parse_loads(self.cim, circuit, busbar_dict)
        self.parse_shunts(self.cim, circuit, busbar_dict)
        self.parse_generators(self.cim, circuit, busbar_dict)

        self.emit_text('Done!')
        return circuit


if __name__ == '__main__':
    import os

    # folder = r'C:\Users\penversa\Documents\Grids\CGMES\TYNDP_2025'
    folder = '/home/santi/Documentos/Private_Grids/CGMES/TYNDP_2025'

    files = [
             # '2025NT_FR_model_004.zip',
             # '2025NT_ES_model_003.zip',
             '2025NT_PT_model_003.zip',
             '20191017T0918Z_ENTSO-E_BD_1130.zip'
             ]
    fnames = [os.path.join(folder, f) for f in files]

    print('Reading...')
    parser = CIMImport(text_func=print)
    circuit_ = parser.load_cim_file(fnames)
    print()
    # parser.cim.to_excel('Spain_data.xlsx')



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


from PySide2 import QtCore
from GridCal.Engine.Devices.wire import Wire
from GridCal.Engine.Devices.tower import Tower, WireInTower

"""
Equations source:
a) ATP-EMTP theory book

Typical values of earth 
10 Ω/m3 - Resistivity of swampy ground 
100 Ω/m3 - Resistivity of average damp earth 
1000 Ω/m3 - Resistivity of dry earth 
"""


class WiresTable(QtCore.QAbstractTableModel):

    def __init__(self, parent=None):

        QtCore.QAbstractTableModel.__init__(self, parent)

        self.header = ['Name', 'R (Ohm/km)', 'GMR (m)', 'max current (kA)']

        self.index_prop = {0: 'name', 1: 'r', 2: 'gmr', 3: 'max_current'}

        self.converter = {0: str, 1: float, 2: float, 3: float}

        self.editable = [True, True, True, True]

        self.wires = list()

    def add(self, wire: Wire):
        """
        Add wire
        :param wire:
        :return:
        """
        row = len(self.wires)
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.wires.append(wire)
        self.endInsertRows()

    def delete(self, index):
        """
        Delete wire
        :param index:
        :return:
        """
        row = len(self.wires)
        self.beginRemoveRows(QtCore.QModelIndex(), row - 1, row - 1)
        self.wires.pop(index)
        self.endRemoveRows()

    def is_used(self, name):
        """
        checks if the name is used
        """
        n = len(self.wires)
        for i in range(n-1, -1, -1):
            if self.wires[i].name == name:
                return True
        return False

    def flags(self, index):
        if self.editable[index.column()]:
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.ItemIsEnabled

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.wires)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(self.header)

    def parent(self, index=None):
        return QtCore.QModelIndex()

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                val = getattr(self.wires[index.row()], self.index_prop[index.column()])
                return str(val)
        return None

    def headerData(self, p_int, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self.header[p_int]

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        """
        if self.editable[index.column()]:
            wire = self.wires[index.row()]
            attr = self.index_prop[index.column()]

            if attr == 'tower_name':
                if self.is_used(value):
                    pass
                else:
                    setattr(wire, attr, self.converter[index.column()](value))
            else:
                setattr(wire, attr, self.converter[index.column()](value))

        return True



class WiresCollection(QtCore.QAbstractTableModel):

    def __init__(self, parent=None, wires_in_tower=()):
        """

        :param parent:
        :param wires_in_tower:
        """
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.header = ['name', 'X (m)', 'Y (m)', 'phase']

        self.index_prop = {0: 'name', 1: 'xpos', 2: 'ypos', 3: 'phase'}

        self.converter = {0: str, 1: float, 2: float, 3: float}

        self.editable = [False, True, True, True]

        self.wires_in_tower = list(wires_in_tower)

    def add(self, wire: WireInTower):
        """
        Add wire
        :param wire:
        :return:
        """
        row = len(self.wires_in_tower)
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.wires_in_tower.append(wire)
        self.endInsertRows()

    def delete(self, index):
        """
        Delete wire
        :param index: index of the wire
        :return:
        """
        row = len(self.wires_in_tower)
        self.beginRemoveRows(QtCore.QModelIndex(), row - 1, row - 1)
        self.wires_in_tower.pop(index)
        self.endRemoveRows()

    def is_used(self, name):
        """
        checks if the name is used
        """
        n = len(self.wires_in_tower)
        for i in range(n-1, -1, -1):
            if self.wires_in_tower[i].name == name:
                return True
        return False

    def flags(self, index):
        if self.editable[index.column()]:
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.ItemIsEnabled

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.wires_in_tower)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(self.header)

    def parent(self, index=None):
        return QtCore.QModelIndex()

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                val = getattr(self.wires_in_tower[index.row()], self.index_prop[index.column()])
                return str(val)
        return None

    def headerData(self, p_int, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self.header[p_int]

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        """
        if self.editable[index.column()]:
            wire = self.wires_in_tower[index.row()]
            attr = self.index_prop[index.column()]

            if attr == 'tower_name':
                if self.is_used(value):
                    pass
                else:
                    setattr(wire, attr, self.converter[index.column()](value))
            else:
                setattr(wire, attr, self.converter[index.column()](value))

        return True


class TowerModel(QtCore.QAbstractTableModel):

    def __init__(self, parent=None, edit_callback=None, tower: Tower=None):
        """

        :param parent:
        :param edit_callback:
        :param tower:
        """

        QtCore.QAbstractTableModel.__init__(self)

        if tower is None:
            self.tower = Tower()
        else:
            self.tower = tower

        # other properties
        self.edit_callback = edit_callback

    def __str__(self):
        return self.tower.name

    def add(self, wire: WireInTower):
        """
        Add wire
        :param wire:
        :return:
        """
        row = self.rowCount()
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.tower.wires_in_tower.append(wire)
        self.endInsertRows()

    def delete(self, index):
        """
        Delete wire
        :param index:
        :return:
        """
        row = self.rowCount()
        self.beginRemoveRows(QtCore.QModelIndex(), row - 1, row - 1)
        self.tower.wires_in_tower.pop(index)
        self.endRemoveRows()

    def delete_by_name(self, wire: Wire):
        """
        Delete wire by name
        :param wire: Wire object
        """
        n = len(self.tower.wires_in_tower)
        for i in range(n-1, -1, -1):
            if self.tower.wires_in_tower[i].name == wire.name:
                self.delete(i)

    def is_used(self, wire: Wire):
        """

        :param wire:
        :return:
        """
        n = len(self.tower.wires_in_tower)
        for i in range(n-1, -1, -1):
            if self.tower.wires_in_tower[i].name == wire.name:
                return True

    def flags(self, index):
        """

        :param index:
        :return:
        """
        if self.tower.editable_wire[index.column()]:
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.ItemIsEnabled

    def rowCount(self, parent=None):
        """

        :param parent:
        :return:
        """
        return len(self.tower.wires_in_tower)

    def columnCount(self, parent=None):
        """

        :param parent:
        :return:
        """
        return len(self.tower.header)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """

        :param index:
        :param role:
        :return:
        """
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                val = getattr(self.tower.wires_in_tower[index.row()], self.tower.index_prop[index.column()])
                return str(val)
        return None

    def headerData(self, p_int, orientation, role):
        """

        :param p_int:
        :param orientation:
        :param role:
        :return:
        """
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self.tower.header[p_int]

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        """
        if self.tower.editable_wire[index.column()]:
            wire = self.tower.wires_in_tower[index.row()]
            attr = self.tower.index_prop[index.column()]

            try:
                val = self.tower.converter[index.column()](value)
            except:
                val = 0

            # correct the phase to the correct range
            if attr == 'phase':
                if val < 0 or val > 3:
                    val = 0

            setattr(wire, attr, val)

            if self.edit_callback is not None:
                self.edit_callback()

        return True




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'gui.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from .matplotlibwidget import MatplotlibWidget

from .icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(1183, 675)
        self.gridLayout = QGridLayout(Dialog)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(1, 1, 1, 1)
        self.tabWidget = QTabWidget(Dialog)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.verticalLayout_6 = QVBoxLayout(self.tab_2)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.main_splitter = QSplitter(self.tab_2)
        self.main_splitter.setObjectName(u"main_splitter")
        self.main_splitter.setOrientation(Qt.Horizontal)
        self.frame_8 = QFrame(self.main_splitter)
        self.frame_8.setObjectName(u"frame_8")
        self.frame_8.setFrameShape(QFrame.NoFrame)
        self.frame_8.setFrameShadow(QFrame.Raised)
        self.verticalLayout_5 = QVBoxLayout(self.frame_8)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.frame_5 = QFrame(self.frame_8)
        self.frame_5.setObjectName(u"frame_5")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_5.sizePolicy().hasHeightForWidth())
        self.frame_5.setSizePolicy(sizePolicy)
        self.frame_5.setFrameShape(QFrame.NoFrame)
        self.frame_5.setFrameShadow(QFrame.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame_5)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_9 = QLabel(self.frame_5)
        self.label_9.setObjectName(u"label_9")

        self.horizontalLayout.addWidget(self.label_9)

        self.name_lineEdit = QLineEdit(self.frame_5)
        self.name_lineEdit.setObjectName(u"name_lineEdit")

        self.horizontalLayout.addWidget(self.name_lineEdit)


        self.verticalLayout_5.addWidget(self.frame_5)

        self.frame_6 = QFrame(self.frame_8)
        self.frame_6.setObjectName(u"frame_6")
        sizePolicy.setHeightForWidth(self.frame_6.sizePolicy().hasHeightForWidth())
        self.frame_6.setSizePolicy(sizePolicy)
        self.frame_6.setFrameShape(QFrame.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_3 = QHBoxLayout(self.frame_6)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_2)

        self.label_8 = QLabel(self.frame_6)
        self.label_8.setObjectName(u"label_8")

        self.horizontalLayout_3.addWidget(self.label_8)

        self.frequency_doubleSpinBox = QDoubleSpinBox(self.frame_6)
        self.frequency_doubleSpinBox.setObjectName(u"frequency_doubleSpinBox")
        self.frequency_doubleSpinBox.setDecimals(0)
        self.frequency_doubleSpinBox.setValue(50.000000000000000)

        self.horizontalLayout_3.addWidget(self.frequency_doubleSpinBox)

        self.label_11 = QLabel(self.frame_6)
        self.label_11.setObjectName(u"label_11")

        self.horizontalLayout_3.addWidget(self.label_11)

        self.rho_doubleSpinBox = QDoubleSpinBox(self.frame_6)
        self.rho_doubleSpinBox.setObjectName(u"rho_doubleSpinBox")
        self.rho_doubleSpinBox.setMaximum(9999999.000000000000000)
        self.rho_doubleSpinBox.setValue(100.000000000000000)

        self.horizontalLayout_3.addWidget(self.rho_doubleSpinBox)


        self.verticalLayout_5.addWidget(self.frame_6)

        self.splitter = QSplitter(self.frame_8)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setMaximumSize(QSize(16777215, 16777215))
        self.splitter.setOrientation(Qt.Vertical)
        self.frame_3 = QFrame(self.splitter)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setFrameShape(QFrame.NoFrame)
        self.frame_3.setFrameShadow(QFrame.Raised)
        self.verticalLayout_8 = QVBoxLayout(self.frame_3)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.label_12 = QLabel(self.frame_3)
        self.label_12.setObjectName(u"label_12")

        self.verticalLayout_8.addWidget(self.label_12)

        self.wires_tableView = QTableView(self.frame_3)
        self.wires_tableView.setObjectName(u"wires_tableView")

        self.verticalLayout_8.addWidget(self.wires_tableView)

        self.frame_7 = QFrame(self.frame_3)
        self.frame_7.setObjectName(u"frame_7")
        self.frame_7.setFrameShape(QFrame.StyledPanel)
        self.frame_7.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_4 = QHBoxLayout(self.frame_7)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.add_to_tower_pushButton = QPushButton(self.frame_7)
        self.add_to_tower_pushButton.setObjectName(u"add_to_tower_pushButton")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/plus.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.add_to_tower_pushButton.setIcon(icon)

        self.horizontalLayout_4.addWidget(self.add_to_tower_pushButton)

        self.horizontalSpacer_3 = QSpacerItem(990, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_3)


        self.verticalLayout_8.addWidget(self.frame_7)

        self.splitter.addWidget(self.frame_3)
        self.frame_4 = QFrame(self.splitter)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setFrameShape(QFrame.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Raised)
        self.verticalLayout_4 = QVBoxLayout(self.frame_4)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(9, 9, 9, 9)
        self.label_10 = QLabel(self.frame_4)
        self.label_10.setObjectName(u"label_10")

        self.verticalLayout_4.addWidget(self.label_10)

        self.tower_tableView = QTableView(self.frame_4)
        self.tower_tableView.setObjectName(u"tower_tableView")

        self.verticalLayout_4.addWidget(self.tower_tableView)

        self.frame = QFrame(self.frame_4)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.delete_from_tower_pushButton = QPushButton(self.frame)
        self.delete_from_tower_pushButton.setObjectName(u"delete_from_tower_pushButton")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/minus.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.delete_from_tower_pushButton.setIcon(icon1)

        self.horizontalLayout_2.addWidget(self.delete_from_tower_pushButton)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.compute_pushButton = QPushButton(self.frame)
        self.compute_pushButton.setObjectName(u"compute_pushButton")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/calc.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.compute_pushButton.setIcon(icon2)
        self.compute_pushButton.setIconSize(QSize(16, 16))

        self.horizontalLayout_2.addWidget(self.compute_pushButton)


        self.verticalLayout_4.addWidget(self.frame)

        self.splitter.addWidget(self.frame_4)

        self.verticalLayout_5.addWidget(self.splitter)

        self.main_splitter.addWidget(self.frame_8)
        self.PlotFrame = QFrame(self.main_splitter)
        self.PlotFrame.setObjectName(u"PlotFrame")
        self.PlotFrame.setFrameShape(QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QFrame.Raised)
        self.verticalLayout_7 = QVBoxLayout(self.PlotFrame)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(9, 9, 9, 9)
        self.label_4 = QLabel(self.PlotFrame)
        self.label_4.setObjectName(u"label_4")

        self.verticalLayout_7.addWidget(self.label_4)

        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName(u"plotwidget")

        self.verticalLayout_7.addWidget(self.plotwidget)

        self.frame_9 = QFrame(self.PlotFrame)
        self.frame_9.setObjectName(u"frame_9")
        self.frame_9.setMaximumSize(QSize(16777215, 24))
        self.frame_9.setFrameShape(QFrame.StyledPanel)
        self.frame_9.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_5 = QHBoxLayout(self.frame_9)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.horizontalSpacer_4 = QSpacerItem(19, 19, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_4)

        self.acceptButton = QPushButton(self.frame_9)
        self.acceptButton.setObjectName(u"acceptButton")

        self.horizontalLayout_5.addWidget(self.acceptButton)


        self.verticalLayout_7.addWidget(self.frame_9)

        self.main_splitter.addWidget(self.PlotFrame)

        self.verticalLayout_6.addWidget(self.main_splitter)

        self.tabWidget.addTab(self.tab_2, "")
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.verticalLayout_3 = QVBoxLayout(self.tab)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.frame_10 = QFrame(self.tab)
        self.frame_10.setObjectName(u"frame_10")
        self.frame_10.setFrameShape(QFrame.StyledPanel)
        self.frame_10.setFrameShadow(QFrame.Raised)
        self.gridLayout_2 = QGridLayout(self.frame_10)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.label_2 = QLabel(self.frame_10)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout_2.addWidget(self.label_2, 0, 1, 1, 1)

        self.label_6 = QLabel(self.frame_10)
        self.label_6.setObjectName(u"label_6")

        self.gridLayout_2.addWidget(self.label_6, 2, 0, 1, 1)

        self.z_tableView_abcn = QTableView(self.frame_10)
        self.z_tableView_abcn.setObjectName(u"z_tableView_abcn")

        self.gridLayout_2.addWidget(self.z_tableView_abcn, 1, 0, 1, 1)

        self.y_tableView_abcn = QTableView(self.frame_10)
        self.y_tableView_abcn.setObjectName(u"y_tableView_abcn")

        self.gridLayout_2.addWidget(self.y_tableView_abcn, 1, 1, 1, 1)

        self.label_7 = QLabel(self.frame_10)
        self.label_7.setObjectName(u"label_7")

        self.gridLayout_2.addWidget(self.label_7, 4, 0, 1, 1)

        self.z_tableView_abc = QTableView(self.frame_10)
        self.z_tableView_abc.setObjectName(u"z_tableView_abc")

        self.gridLayout_2.addWidget(self.z_tableView_abc, 3, 0, 1, 1)

        self.label = QLabel(self.frame_10)
        self.label.setObjectName(u"label")

        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)

        self.z_tableView_seq = QTableView(self.frame_10)
        self.z_tableView_seq.setObjectName(u"z_tableView_seq")

        self.gridLayout_2.addWidget(self.z_tableView_seq, 5, 0, 1, 1)

        self.label_3 = QLabel(self.frame_10)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout_2.addWidget(self.label_3, 2, 1, 1, 1)

        self.y_tableView_abc = QTableView(self.frame_10)
        self.y_tableView_abc.setObjectName(u"y_tableView_abc")

        self.gridLayout_2.addWidget(self.y_tableView_abc, 3, 1, 1, 1)

        self.label_5 = QLabel(self.frame_10)
        self.label_5.setObjectName(u"label_5")

        self.gridLayout_2.addWidget(self.label_5, 4, 1, 1, 1)

        self.y_tableView_seq = QTableView(self.frame_10)
        self.y_tableView_seq.setObjectName(u"y_tableView_seq")

        self.gridLayout_2.addWidget(self.y_tableView_seq, 5, 1, 1, 1)


        self.verticalLayout_3.addWidget(self.frame_10)

        self.tabWidget.addTab(self.tab, "")

        self.gridLayout.addWidget(self.tabWidget, 4, 0, 1, 1)


        self.retranslateUi(Dialog)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Tower creation", None))
        self.label_9.setText(QCoreApplication.translate("Dialog", u"Name", None))
        self.label_8.setText(QCoreApplication.translate("Dialog", u"Frequency (Hz)", None))
        self.label_11.setText(QCoreApplication.translate("Dialog", u"Earth resistivity (Ohm/m^3)", None))
        self.label_12.setText(QCoreApplication.translate("Dialog", u"Wire catalogue", None))
#if QT_CONFIG(tooltip)
        self.add_to_tower_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Add wire", None))
#endif // QT_CONFIG(tooltip)
        self.add_to_tower_pushButton.setText("")
        self.label_10.setText(QCoreApplication.translate("Dialog", u"Wire compisition", None))
#if QT_CONFIG(tooltip)
        self.delete_from_tower_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Delete wire", None))
#endif // QT_CONFIG(tooltip)
        self.delete_from_tower_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.compute_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Compute matrices", None))
#endif // QT_CONFIG(tooltip)
        self.compute_pushButton.setText("")
        self.label_4.setText(QCoreApplication.translate("Dialog", u"Tower", None))
        self.acceptButton.setText(QCoreApplication.translate("Dialog", u"Accept", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QCoreApplication.translate("Dialog", u"Tower designer", None))
        self.label_2.setText(QCoreApplication.translate("Dialog", u"   Y shunt (uS / km) for ABCN", None))
        self.label_6.setText(QCoreApplication.translate("Dialog", u"   Z series (Ohm / km) for ABC", None))
        self.label_7.setText(QCoreApplication.translate("Dialog", u"   Z series (Ohm / km) in sequence components", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"   Z series (Ohm / km) for ABCN", None))
        self.label_3.setText(QCoreApplication.translate("Dialog", u"   Y shunt (uS / km) for ABC", None))
        self.label_5.setText(QCoreApplication.translate("Dialog", u"   Y shunt (uS / km) for the sequence components", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("Dialog", u"Impedance matrices", None))
    # retranslateUi




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import numpy as np
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.results_table import ResultsTable
from GridCal.Engine.Simulations.results_template import ResultsTemplate


class ContingencyAnalysisResults(ResultsTemplate):

    def __init__(self, nbus, nbr, bus_names, branch_names, bus_types):
        """
        TimeSeriesResults constructor
        @param nbus: number of buses
        @param nbr: number of branches
        """
        ResultsTemplate.__init__(self,
                                 name='Contingency Analysis Results',
                                 available_results=[ResultTypes.OTDF,
                                                    ResultTypes.BusActivePower,
                                                    ResultTypes.BranchActivePowerFrom,
                                                    ResultTypes.BranchLoading],
                                 data_variables=['bus_types',
                                                 'branch_names',
                                                 'bus_names',
                                                 'voltage',
                                                 'S',
                                                 'Sf',
                                                 'loading',
                                                 'otdf'])

        self.branch_names = branch_names

        self.bus_names = bus_names

        self.bus_types = bus_types

        self.voltage = np.ones((nbr, nbus), dtype=complex)

        self.S = np.zeros((nbr, nbus), dtype=complex)

        self.Sf = np.zeros((nbr, nbr), dtype=complex)

        self.loading = np.zeros((nbr, nbr), dtype=complex)

        self.otdf = np.zeros((nbr, nbr))

    def apply_new_rates(self, nc: "SnapshotData"):
        rates = nc.Rates
        self.loading = self.Sf / (rates + 1e-9)

    def get_steps(self):
        return

    def get_results_dict(self):
        """
        Returns a dictionary with the results sorted in a dictionary
        :return: dictionary of 2D numpy arrays (probably of complex numbers)
        """
        data = {'Vm': np.abs(self.voltage).tolist(),
                'Va': np.angle(self.voltage).tolist(),
                'P': self.S.real.tolist(),
                'Q': self.S.imag.tolist(),
                'Sbr_real': self.Sf.real.tolist(),
                'Sbr_imag': self.Sf.imag.tolist(),
                'loading': np.abs(self.loading).tolist()}
        return data

    def mdl(self, result_type: ResultTypes):
        """
        Plot the results
        :param result_type:
        :return:
        """

        index = self.branch_names

        if result_type == ResultTypes.BusVoltageModule:
            data = np.abs(self.voltage)
            y_label = '(p.u.)'
            title = 'Bus voltage '
            labels = self.bus_names
            # index = self.branch_names

        elif result_type == ResultTypes.BusVoltageAngle:
            data = np.angle(self.voltage, deg=True)
            y_label = '(Deg)'
            title = 'Bus voltage '
            labels = self.bus_names
            # index = self.branch_names

        elif result_type == ResultTypes.BusActivePower:
            data = self.S.real
            y_label = '(MW)'
            title = 'Bus active power '
            labels = self.bus_names
            # index = self.branch_names

        elif result_type == ResultTypes.BranchActivePowerFrom:
            data = self.Sf.real
            y_label = 'MW'
            title = 'Branch active power '
            labels = ['# ' + x for x in self.branch_names]
            # index = self.branch_names

        elif result_type == ResultTypes.BranchLoading:
            data = self.loading.real * 100
            y_label = '(%)'
            title = 'Branch loading '
            labels = ['# ' + x for x in self.branch_names]
            # index = self.branch_names

        elif result_type == ResultTypes.OTDF:
            data = self.otdf
            y_label = 'Per unit'
            labels = ['# ' + x for x in self.branch_names]

            title = 'LODF'
        else:
            raise Exception('Result type not understood:' + str(result_type))

        # assemble model
        mdl = ResultsTable(data=data,
                           index=index,
                           columns=labels,
                           title=title,
                           ylabel=y_label)
        return mdl

